"""
Items used to support saving classes to HDF 5 files.
"""
import h5py
from abc import ABC, abstractmethod
from typing import get_type_hints, List, Union
from inspect import isclass
from numpy import ndarray
from logging import debug, warning

try:
    from typing import get_args, get_origin
except ImportError:
    def get_args(tp): return getattr(tp, '__args__', ())
    def get_origin(tp): return getattr(tp, '__extra__', None)

class H5Savable(ABC):
    # Defines minimum required methods for a class to be used upon saving
    @property
    @abstractmethod
    def h5group_name(self):
        """
        Attribute that stores the name of the group it will be supplied in the write_/read_h5group methods.
        The main reason to fix this in the H5Savable class is to make it more transparent, where this
        classes attributes will be saved in the hdf5 file.
        """

    @abstractmethod
    def write_h5group(self, group:h5py.Group):
        """ Save configuration to hdf5 group """

    @abstractmethod
    def read_h5group(self, group:h5py.Group):
        """ Configure object from hdf5 group """

    def h5_write_free_dict(self, group:h5py.Group, obj:dict):
        """ Help method to write an arbitrary dictionary of variable depth to hdf5 group """
        for key, value in obj.items():
            vtyp=type(value)
            if vtyp is dict:
                sub_group=group.create_group(key)
                self.h5_write_free_dict(sub_group, value)
            elif any([issubclass(vtyp, typ) for typ in [float, int, ndarray]]):
                group[key] = value
            elif issubclass(vtyp, str):
                group[key]=value.encode('utf-8')

    def h5_read_free_dict(self, output:dict, group:Union[h5py.Group,h5py.Dataset], item_path:List[str]):
        """
        Recursive read of meta data from hdf5 group
        """
        node=output
        for pathi in item_path[:-1]:
            # decent into sub-node, create if it does not exist
            if pathi in node:
                node=node[pathi]
            else:
                new_node={}
                node[pathi]=new_node
                node=new_node
        if type(group) is h5py.Dataset:
            value=group[()]
            vtyp=type(value)
            if issubclass(vtyp, float):
                value=float(value)
            elif issubclass(vtyp, int):
                value=int(value)
            elif vtyp is bytes:
                value=value.decode('utf-8')
            node[item_path[-1]]=value
            return
        for key in group:
            self.h5_read_free_dict(output, group[key], item_path+[key])


# noinspection PyAbstractClass
class H5HintedExport(H5Savable):
    """
    Base class that exports all its attributes that have a type hint to the hdf5 group using their name as key.

    This allows clean classes where parameters are automatically written correctly and type checked.
    """
    _export_ignore=[] # allows type hinted parameters that should not be exported

    def init_defaults(self):
        """Allows the class to automatically initialize the exported parameters from their defaults"""
        for attr, typ in get_type_hints(self).items():
            if attr in self._export_ignore or attr.startswith('_'):
                continue
            if not hasattr(self.__class__, attr):
                continue
            default=getattr(self.__class__, attr)
            if hasattr(default, 'copy'):
                default = default.copy()  # for array, dict and lists, make sure a copy is used
            setattr(self, attr, default)

    def write_h5group(self, group:h5py.Group):
        """
        Uses this objects type hints to write attributes to hdf5 group.

        Two object types are treated specifically:
            -dict will be written in sub-group items
            -H5Savable derived classes will be written in sub-group
             using their write_h5group method and h5group_name attribute
        """
        for attr, typ in get_type_hints(self).items():
            if attr in self._export_ignore or attr.startswith('_'):
                continue
            value=getattr(self, attr)
            if typ is dict:
                # free dictionary, save every str, int, float type values and ignore rest
                self.h5_write_free_dict(group.create_group(attr), value)
            elif get_origin(typ) is dict:
                sub_group=group.create_group(attr)
                styp=get_args(typ)[1]
                for key, subval in value.items():
                    if styp is str:
                        sub_group[key] = subval.encode('utf-8')
                    else:
                        sub_group[key] = subval
            elif isclass(typ) and issubclass(typ, H5Savable):
                # the attribute is savable, create a sub-group
                sub_group = group.create_group(value.h5group_name)
                value.write_h5group(sub_group)
            else:
                if typ is str:
                    value=value.encode('utf-8')
                group[attr] = value

    def read_h5group(self, group:h5py.Group):
        """
        Uses this objects type hints to read attributes from hdf5 group.

        Two object types are treated specifically:
            -dict will be read from sub-group items
            -H5Savable derived classes will be read from sub-group
             using their read_h5group method and h5group_name attribute

        Values missing in file are handled according to two cases:
            1. If the class itself defines values for the parameters, these are taken as
               fallback and only debug info is provided
            2. Otherwise a warning message for missing parameter is given and attribute is not set at all
        """
        for attr, typ in get_type_hints(self).items():
            if attr in self._export_ignore or attr.startswith('_'):
                continue
            if hasattr(self.__class__, attr):
                fallback=getattr(self.__class__, attr)
            else:
                fallback=None
            write_attr=None
            err_mesg=''

            if typ is dict:
                opt = {}
                try:
                    subgroup=group[attr]
                except KeyError:
                    warning(f'Did not find hdf5 group for {attr} in {group.name}')
                    continue
                self.h5_read_free_dict(opt, subgroup, [])
                write_attr=opt
            elif get_origin(typ) is dict:
                try:
                    subgroup=group[attr]
                except KeyError:
                    warning(f'Did not find hdf5 group for {attr} in {group.name}')
                    continue
                opt={}
                vtyp=get_args(typ)[1]
                for key in subgroup:
                    if vtyp is ndarray:
                        opt[key]=subgroup[key][()]
                    elif vtyp is str:
                        value= subgroup[key][()]
                        if type(value) is str:
                            opt[key]=value
                        else:
                            try:
                                opt[key] =value.decode('utf-8')
                            except ValueError:
                                warning(f'Could not convert {attr}/{key} in {group.name} to string')
                                continue
                    else:
                        try:
                            opt[key]=vtyp(subgroup[key][()])
                        except ValueError:
                            warning(f'Could not convert {attr}/{key} in {group.name} to {vtyp.__name__}')
                            continue
                write_attr=opt
            elif isclass(typ) and issubclass(typ, H5Savable):
                # the attribute is loadable from sub-group
                obj=getattr(self, attr)
                try:
                    subgroup = group[obj.h5group_name]
                except KeyError:
                    warning(f'Did not find hdf5 group for {attr} in {group.name}')
                    continue
                obj.read_h5group(subgroup)
                continue
            else:
                try:
                    value = group[attr][()]
                except KeyError: # resulting item did not exist
                    err_mesg=f'\n    Reason: The entry with name {attr} did not exist.'
                except AttributeError: # result was not a dataset
                    err_mesg=f'\n    Reason: The item with name {attr} was not a Dataset but {type(group[attr])}.'
                else:
                    if get_origin(typ) is tuple:
                        valtyp = get_args(typ)
                        try:
                            write_attr=tuple(ti(vi) for ti,vi in zip(valtyp, value))
                        except ValueError:
                            err_mesg = f'\n    Reason: Could not convert value {value} to types {valtyp}.'
                    elif typ is ndarray:
                        write_attr=value
                    elif typ is str and type(value) is bytes:
                        try:
                            write_attr=value.decode('utf-8')
                        except UnicodeDecodeError:
                            err_mesg = f'\n    Reason: Could not convert bytes to string using "utf-8" encoding.'
                    else:
                        try:
                            write_attr=typ(value)
                        except ValueError:
                            err_mesg = f'\n    Reason: Could not convert value to type {typ.__name__}.'

            if write_attr is not None:
                setattr(self, attr, write_attr)
            elif fallback is not None:
                debug(f'Could not load {attr} from {group.name}, setting fallback value.'+err_mesg)
                if hasattr(fallback, 'copy'):
                    fallback=fallback.copy() # for array, dict and lists, make sure a copy is used
                setattr(self, attr, fallback)
            else:
                warning(f'Could not load {attr} from {group.name} and no fallback was available, ignoring item.\n'
                        f'(This may result in strange behavior as the model will have values from before the load.)'+
                        err_mesg)




"""
A general wx.Dialog to deal with user configurable settings.
"""

import wx
from math import log10

from ..core.config import BaseConfig

class SettingsDialog(wx.Dialog):
    def __init__(self, parent: wx.Window, settings: BaseConfig, title='Settings', apply_callback:callable=None):
        '''
        Configuration dialog for BaseConfig based classes.

        The apply_callback is a function that is called with a dictionary of updated options that
        can be evaluated. It returns a bool to determin if the dialog should apply these options to the
        config object.
        '''
        wx.Dialog.__init__(self, parent, -1, title, style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)
        self.settings=settings
        if apply_callback is None:
            self.apply_callback=lambda options: True
        else:
            self.apply_callback=apply_callback

        self.build_sizers()
        self.rb_actions={} # entries for toggling radio button selected entries enabled
        self.rb_bool_fields={} # the associated with the above action
        self.value_entries={} # all controls for settable values, the key is field.name

        groups=list(self.settings.groups.keys())
        n_groups=len(groups)
        r_groups=n_groups//2
        l_groups=n_groups-r_groups
        for lg in groups[:l_groups]:
            self.build_group(lg, self.row_sizer1)
        for rg in groups[l_groups:]:
            self.build_group(rg, self.row_sizer2)

        self.GetSizer().Fit(self)
        self.Layout()

    def build_sizers(self, add_apply=False):
        col_sizer=wx.BoxSizer(wx.HORIZONTAL)
        self.row_sizer1=wx.BoxSizer(wx.VERTICAL)
        self.row_sizer2=wx.BoxSizer(wx.VERTICAL)
        col_sizer.Add(self.row_sizer1, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)
        col_sizer.Add(self.row_sizer2, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)

        # Add the Dialog buttons
        button_sizer=wx.StdDialogButtonSizer()
        okay_button=wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        if add_apply:
            apply_button=wx.Button(self, wx.ID_APPLY)
            apply_button.SetDefault()
            button_sizer.AddButton(apply_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))
        button_sizer.Realize()
        # Add some event handlers
        self.Bind(wx.EVT_BUTTON, self.on_apply_changes)

        sizer=wx.BoxSizer(wx.VERTICAL)
        sizer.Add(col_sizer, 1, wx.GROW|wx.LEFT|wx.RIGHT, 10)
        line=wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.TOP, 10)

        sizer.Add(button_sizer, 0, flag=wx.ALIGN_RIGHT, border=4)
        self.SetSizer(sizer)


    def build_group(self, name:str, sizer:wx.BoxSizer):
        box=wx.StaticBox(self, -1, name)
        box_sizer=wx.StaticBoxSizer(box, orient=wx.VERTICAL)
        sizer.Add(box_sizer, 0, wx.EXPAND)

        pnames=self.settings.groups[name]
        fields=dict((fi.name, fi) for fi in self.settings.get_fields())
        for ni in pnames:
            if type(ni) in [list, tuple]:
                if len(ni)==4 and fields[ni[1]].type is bool:
                    # the entry is toggling type
                    e1_active=fields[ni[1]]
                    e1_field=fields[ni[2]]
                    e2_field=fields[ni[3]]
                    gbox=wx.StaticBox(box, -1, ni[0])
                    gbox_sizer = wx.StaticBoxSizer(gbox, orient=wx.VERTICAL)
                    box_sizer.Add(gbox_sizer, 0, wx.EXPAND)
                    self.rb_bool_fields[gbox.GetId()]=ni[1]
                    e1=self.get_toggle_entry(e1_field, gbox, getattr(self.settings, e1_active.name, True))
                    gbox_sizer.Add(e1, 1, wx.EXPAND|wx.ALL, 2)
                    e2=self.get_toggle_entry(e2_field, gbox, not getattr(self.settings, e1_active.name, True))
                    gbox_sizer.Add(e2, 1, wx.EXPAND|wx.BOTTOM, 5)
                    gbox.Bind(wx.EVT_RADIOBUTTON, self.on_radio)
                else:
                    hsizer=wx.BoxSizer(wx.HORIZONTAL)
                    box_sizer.Add(hsizer, 0, wx.ALIGN_CENTRE|wx.ALL, 5)
                    for nj in ni:
                        field = fields[nj]
                        entry = self.get_entry(field, box)
                        hsizer.Add(entry, 0, wx.EXPAND | wx.ALL, 2)
            else:
                field=fields[ni]
                entry=self.get_entry(field, box)
                box_sizer.Add(entry, 1, wx.FIXED_MINSIZE|wx.ALIGN_CENTRE|wx.ALL, 5)
        box_sizer.Fit(box)

    def get_entry(self, field, parent, add_label=True):
        gmeta=field.metadata.get('genx', {})
        if 'label' in gmeta:
            name=gmeta['label']
        else:
            name=field.name.replace('_', ' ')
        if field.type is bool:
            entry=wx.CheckBox(parent, -1, name, style=wx.ALIGN_CENTER)
            entry.SetValue(getattr(self.settings, field.name, False))
            self.value_entries[field.name]=entry
        elif 'selection' in gmeta:
            choice = wx.Choice(parent, -1, choices=gmeta['selection'])
            choice.SetSelection(gmeta['selection'].index(getattr(self.settings, field.name, gmeta['selection'][0])))
            self.value_entries[field.name]=choice
            entry = self.label_entry(parent, choice, name)
        elif 'pmin' in gmeta and 'pmax' in gmeta:
            # a numerical parameter with given range
            if field.type is int:
                spin=wx.SpinCtrl(parent, -1, min=gmeta['pmin'], max=gmeta['pmax'])
            else:
                inc=10**(round(log10(gmeta['pmax']))-2)
                spin = wx.SpinCtrlDouble(parent, -1, min=gmeta['pmin'], max=gmeta['pmax'], inc=inc)
                spin.SetDigits(4)
            spin.SetValue(getattr(self.settings, field.name, gmeta['pmin']))

            spin.SetMinSize(spin.GetSizeFromTextSize(spin.GetTextExtent("%.4f"%gmeta['pmax']+" ")))
            self.value_entries[field.name]=spin
            if add_label:
                entry = self.label_entry(parent, spin, name)
            else:
                entry = spin
        else:
            txt=wx.TextCtrl(parent, -1, style=0)
            txt.SetValue(str(getattr(self.settings, field.name, '')))
            txt.SetMinSize(txt.GetSizeFromTextSize(txt.GetTextExtent(10*' ')))
            self.value_entries[field.name]=txt
            if add_label:
                entry = self.label_entry(parent, txt, name)
            else:
                entry=txt
        return entry

    def get_toggle_entry(self, field, parent, active):
        gmeta=field.metadata.get('genx', {})
        if 'label' in gmeta:
            name=gmeta['label']
        else:
            name=field.name.replace('_', ' ')
        # generate an entries with a radio button to toggle it active
        entry = wx.BoxSizer(wx.HORIZONTAL)
        entry_active=wx.RadioButton(parent, -1, name, style=wx.ALIGN_RIGHT)
        entry_active.SetValue(active)
        entry.Add(entry_active, 1, wx.FIXED_MINSIZE|wx.RIGHT, 8)
        field_entry=self.get_entry(field, parent, add_label=False)
        field_entry.Enable(active)
        parent_id=parent.GetId()
        if parent_id in self.rb_actions:
            self.rb_actions[parent_id][entry_active.GetId()]=field_entry
        else:
            self.rb_actions[parent_id]={entry_active.GetId(): field_entry}
        entry.Add(field_entry, 1, wx.FIXED_MINSIZE, 0)
        return entry

    def on_radio(self, event:wx.CommandEvent):
        # activates the entry next to the selected radio button and deactivates all others in that group
        parent_id=event.GetEventObject().GetParent().GetId()
        sender_id=event.GetId()
        for id, field_entry in self.rb_actions[parent_id].items():
            field_entry.Enable(sender_id==id)
        event.Skip()

    def label_entry(self, parent, entry, label):
        out_entry = wx.BoxSizer(wx.HORIZONTAL)
        out_entry.Add(wx.StaticText(parent, -1, label, style=wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL),
                      0, wx.FIXED_MINSIZE|wx.ALIGN_CENTER_VERTICAL | wx.RIGHT, 8)
        out_entry.Add(entry, 0, wx.EXPAND, 0)
        return out_entry

    def collect_results(self):
        # create a dictionary with all values from the dialog entry
        out={}
        fields=dict((fi.name, fi) for fi in self.settings.get_fields())
        for key, entry in self.value_entries.items():
            field=fields[key]
            if hasattr(entry, 'GetValue'):
                try:
                    out[key]=field.type(entry.GetValue())
                except ValueError:
                    # TODO: add handling of entry errors
                    pass
            elif hasattr(entry, 'GetString') and hasattr(entry, 'GetSelection'):
                out[key]=entry.GetString(entry.GetSelection())
            else:
                raise NotImplementedError(f"Could not evaluate wx Control {entry!r}")
        for pid, rbg in self.rb_actions.items():
            # get the active state of the boolean fields for radio buttons
            out[self.rb_bool_fields[pid]]=list(rbg.values())[0].IsEnabled()
        return out

    def on_apply_changes(self, event):
        event.Skip()
        options=self.collect_results()
        OK_pressed=event.GetId()==wx.ID_OK
        if self.apply_callback(options) and OK_pressed:
            for key, value in options.items():
                setattr(self.settings, key, value)



''' Module to provide the Layer - Stack - Sample classes to build a sample for reflectivity modelling.

Classes:
ReflFunction - A function class that can be used as a parameter in the other classes.
is_reflfunction - Funtion that checks if an object belongs to the class ReflFunction
ReflBase - Base class for all the physical classes.
LayerBase - Base Layer class.
StackBase - Base Stack class.
SampleBase - Base Sample class.
InstrumentBase - Base Instrument class.
ModelFactory - A factory class that creates specific instances of the Base classes for a certain set of parameters.
MakeClasses - Function that creates the classes (backward comparability).
'''

import numpy as np
from genx.core.custom_logging import iprint

class ReflFunction:
    def __init__(self, function, validation_args, validation_kwargs, id=None):
        """ Creates the Refl Function given function. The arguments validation_args and
        validation_kwargs will be used to the validate the returned type of the function by passing
        them to function. The variable id should be a unique string to identify the type ReflFunction.
        """
        self.__func__=function
        self.validation_args=validation_args
        self.validation_kwargs=validation_kwargs
        self.id=id

    def __call__(self, *args, **kwargs):
        return self.__func__(*args, **kwargs)

    def validate(self):
        ''' Function to test that the function returns the anticipated type
        '''
        return self.__call__(*self.validation_args, **self.validation_kwargs)

    def _check_obj(self, other):
        """ Checks the object other so that it fulfills the demands for arithmetic operations.
        """
        supported_types=[int, float, int, complex, np.float64, np.float32]
        if is_reflfunction(other):
            if self.id!=other.id:
                raise TypeError("Two ReflFunction objects must have identical id's to conduct arithmetic operations")
        elif not type(other) in supported_types:
            raise TypeError("%s is not supported for arithmetic operations "%(repr(type(other)))+
                            "of a ReflFunction. It has to int, float, long or complex")

    def __mul__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)*other(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)*other
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rmul__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return other(*args, **kwargs)*self(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return other*self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __add__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)+other(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)+other
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __radd__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return other(*args, **kwargs)+self(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return other+self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __sub__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)-other(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)-other
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rsub__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return other(*args, **kwargs)-self(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return other-self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __div__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)/other(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)/other
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rdiv__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return other(*args, **kwargs)/self(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return other/self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __neg__(self):
        def new_func(*args, **kwargs):
            return -self(*args, **kwargs)

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __pos__(self):
        def new_func(*args, **kwargs):
            return self(*args, **kwargs)

        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __pow__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)**other(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return self(*args, **kwargs)**other
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

    def __rpow__(self, other):
        self._check_obj(other)
        if is_reflfunction(other):
            def new_func(*args, **kwargs):
                return other(*args, **kwargs)**self(*args, **kwargs)
        else:
            def new_func(*args, **kwargs):
                return other**self(*args, **kwargs)
        return ReflFunction(new_func, self.validation_args, self.validation_kwargs, self.id)

def cast_to_array(list_of_obj, *args, **kwargs):
    ''' Casts an list_of_obj, can be a number or an ReflFunction, into an list of evaluated values'''
    id=''
    shape=False
    ret_list=[]
    for obj in list_of_obj:
        if is_reflfunction(obj):
            if id=='':
                id=obj.id
                ret_list.append(obj(*args, **kwargs))
                # Check if we have got an array
                if not np.isscalar(ret_list[-1]):
                    shape=ret_list[-1].shape
            elif id==obj.id:
                ret_list.append(obj(*args, **kwargs))
            else:
                TypeError("Two ReflFunction objects must have identical id's in order to merge them into an array")
        else:
            # We assume that this is an object that can be transformed into an array later on.
            ret_list.append(obj)
    # if we have an array make sure that all the objects have the same shape
    if shape:
        base_array=np.ones(shape)
        nret_list=[]
        for item in ret_list:
            if np.isscalar(item):
                nret_list.append(item*base_array)
            else:
                nret_list.append(item)
        ret_list=nret_list

    return np.array(ret_list)

def harm_sizes(ar, shape, dtype=np.float64):
    '''Utility function to add an additional axis if needed to fulfill the size in shape'''
    ar=np.array(ar, dtype=dtype)
    if shape is None:
        return ar
    elif len(ar.shape)<len(shape):
        return np.array(ar[..., np.newaxis]*np.ones(shape), dtype=dtype)
    elif ar.shape==shape:
        return ar
    else:
        raise TypeError('The size of the array, %s, can not be changed to shape %s'%(ar.shape, shape))

def is_reflfunction(obj):
    ''' Convenience function to determine whether obj belongs to the ReflFunction class.
    Return boolean.
    '''
    return obj.__class__.__name__=='ReflFunction'

def resolve_par(obj, par_name, max_recur=100):
    """ Resolves a layer parameter and takes into account a coupling to another object.
    If obj.par_name is of type obj the relation will be recursive, obj.par_name.par_name.
    A maximum of 100 levels are conducted before an error is thrown.

    :param obj: An object
    :param par_name: A name of a member
    :param max_recur: Maximum number of recursions.
    :return: a value
    """
    val=getattr(obj, par_name)
    i=0
    while type(val) is type(obj) and i<max_recur and not is_reflfunction(val):
        val=getattr(val, par_name)
        i+=1

    return val

class ReflBase:
    # _parameters is a dict of parameter names with their default values
    _parameters={}

    def __init__(self, **kargs):
        # Setup all the parameters in the class
        for par in self._parameters:
            setattr(self, par, self._parameters[par])
            iscomplex=type(self._parameters[par]) is complex
            # Making the set function
            self._make_set_func(par, iscomplex)
            # Creating the get function
            self._make_get_func(par, iscomplex)

        # Set all parameters given as keyword arguments
        for k in kargs:
            if not k in self._parameters and not k in dir(self):
                iprint('%s is not an parameter in %s so it is ignored'%
                       (k, self.__class__.__name__))
            else:
                setattr(self, k, kargs[k])

    def _make_get_func(self, par, iscomplex=False):
        ''' Creates a get function for parameter par and binds it to the object
        '''

        def get_func():
            return getattr(self, par)

        get_func.__name__='get'+par.capitalize()
        setattr(self, get_func.__name__, get_func)

        if iscomplex:
            def get_real_func():
                return getattr(self, par).real

            get_real_func.__name__=get_func.__name__+'real'
            setattr(self, get_real_func.__name__, get_real_func)

            def get_imag_func():
                return getattr(self, par).imag

            get_imag_func.__name__=get_func.__name__+'imag'
            setattr(self, get_imag_func.__name__, get_imag_func)

    def _make_set_func(self, par, iscomplex=False):
        ''' Creates a set function for parameter par and binds it to the object
            '''

        def set_func(val):
            setattr(self, par, val)

        set_func.__name__='set'+par.capitalize()
        setattr(self, set_func.__name__, set_func)

        if iscomplex:
            def set_real_func(val):
                setattr(self, par, val+getattr(self, par).imag*1J)

            set_real_func.__name__=set_func.__name__+'real'
            setattr(self, set_real_func.__name__, set_real_func)

            def set_imag_func(val):
                setattr(self, par, val*1J+getattr(self, par).real)

            set_imag_func.__name__=set_func.__name__+'imag'
            setattr(self, set_imag_func.__name__, set_imag_func)

    def get_parameters(self):
        ''' Returns all the parameters of the current object '''

        return self.parameters

class InstrumentBase(ReflBase):
    _parameters={}

    def __repr__(self):
        s='Instrument('

        for k in self._parameters:
            # if the type is a string...
            if type(getattr(self, k))==type(''):
                stemp="%s='%s', "%(k, str(getattr(self, k)))
            else:
                stemp='%s=%s, '%(k, str(getattr(self, k)))
            s=s+stemp
        return s[:-1]+')'

class LayerBase(ReflBase):
    _parameters={}

    def __repr__(self):
        s='Layer('
        for k in self._parameters:
            s+='%s=%s, '%(k, str(getattr(self, k)))
        return s[:-2]+')'

class StackBase(ReflBase):
    _parameters={'Repetitions': 1}
    Layers=[]
    Repetitions=1

    def __init__(self, **kargs):
        ReflBase.__init__(self, **kargs)
        if 'Layers' in kargs:
            self.Layers=kargs['Layers']
            kargs.pop('Layers')
        else:
            self.Layers=[]

    def __repr__(self):
        s='Stack: '
        for k in self._parameters:
            s+='%s=%s, '%(k, str(getattr(self, k)))
        s=s[:-2]+'\n'
        it=len(self.Layers)
        for lay in range(it-1, -1, -1):
            s+='\t'+repr(self.Layers[lay])+'\n'
        return s

    def resolveLayerParameter(self, parameter):
        par=[resolve_par(lay, parameter)
             for lay in self.Layers]*int(self.Repetitions)
        return par

# Class Sample

class SampleBase(ReflBase):
    _parameters={}
    Ambient=None
    Substrate=None
    Stacks=[]

    def __init__(self, **kargs):
        ReflBase.__init__(self, **kargs)
        special_objects=['Ambient', 'Substrate', 'Stacks']
        for p in special_objects:
            if p in kargs:
                setattr(self, p, kargs[p])
            else:
                raise ValueError('%s has to be defined to create a Sample object'%p)
        # Create a new dictonary for the new arguments
        nkargs={}
        for p in kargs:
            if not p in special_objects:
                nkargs[p]=kargs[p]

    def __repr__(self):
        Add='Sample: '
        for k in self._parameters:
            if type(getattr(self, k))==type(''):
                Add+="%s='%s', "%(k, str(getattr(self, k)))
            else:
                Add+='%s=%s, '%(k, str(getattr(self, k)))
            # Add += '%s=%s, ' % (k, getattr(self, k).__repr__())
        if len(self._parameters)>0:
            Add=Add[:-2]+'\n'
        else:
            Add=Add[:-1]+'\n'
        temp=[repr(x) for x in self.Stacks]
        temp.reverse()
        return (Add+'Ambient:\n\t'+repr(self.Ambient)+'\n'+''.join(temp)
                +'Substrate:\n\t'+repr(self.Substrate))

    def resolveLayerParameters(self):
        par=self.Substrate._parameters.copy()
        for k in par:
            par[k]=[resolve_par(self.Substrate, k)]
        for k in self.Substrate._parameters:
            for stack in self.Stacks:
                par[k]=par[k]+stack.resolveLayerParameter(k)
            par[k]=par[k]+[resolve_par(self.Ambient, k)]
        return par

# Factory needs to:
#            add members and their names in the _parameter list
#            Add set and get methods for each parameter
class ModelFactory:
    ''' Class that creates new classes to implement certain parameters 
        dynamically.
    '''

    def __init__(self):
        class Layer(LayerBase): _parameters=LayerBase._parameters.copy()

        self.Layer=Layer

        class Stack(StackBase): _parameters=StackBase._parameters.copy()

        self.Stack=Stack

        class Sample(SampleBase): _parameters=SampleBase._parameters.copy()

        self.Sample=Sample

        class Instrument(InstrumentBase): _parameters=InstrumentBase._parameters.copy()

        self.Instrument=Instrument

    @staticmethod
    def _add_parameter(cls, par_name, def_val):
        cls._parameters.update({par_name: def_val})

    @staticmethod
    def _add_sim_method(cls, name, func):
        def method(self, *args):
            nargs=args[:-1]+(self,)+(args[-1],)
            return func(*nargs)

        method.__name__='Sim'+name
        setattr(cls, method.__name__, method)

    def set_layer_parameters(self, parameters):
        '''Adds extra parameters (dict with name: value entries) to the
            Layer class'''
        [self._add_parameter(self.Layer, k, parameters[k]) for k in parameters]

    def set_stack_parameters(self, parameters):
        '''Adds extra parameters (dict with name: value entries) to the
            Stack class'''
        # The if not is a hook to so that everything works with the old implementation
        [self._add_parameter(self.Stack, k, parameters[k]) for k in parameters
         if not k in ['Layers', 'Repetitions']]

    def set_sample_parameters(self, parameters):
        '''Adds extra parameters (dict with name: value entries) to the
            Sample class'''
        # The if not is a hook to so that everything works with the old implementation
        [self._add_parameter(self.Sample, k, parameters[k]) for k in parameters
         if not k in ['Ambient', 'Substrate', 'Stacks']]

    def set_simulation_functions(self, functions):
        ''' Adds the simulation functions to the sample class
        '''
        [self._add_sim_method(self.Sample, k, functions[k]) for k in functions]

    def set_instrument_parameters(self, parameters):
        '''Adds extra parameters (dict with name: value entries) to the
            Instrument class'''
        [self._add_parameter(self.Instrument, k, parameters[k]) for k in parameters]

    def get_layer_class(self):
        return self.Layer

    def get_stack_class(self):
        return self.Stack

    def get_sample_class(self):
        return self.Sample

    def get_instrument_class(self):
        return self.Instrument

# Function! to create classes with the model given by the input parameters
def MakeClasses(InstrumentParameters=None,
                LayerParameters=None,
                StackParameters=None,
                SampleParameters=None,
                SimulationFunctions=None,
                ModelID='Standard'):
    if InstrumentParameters is None:
        InstrumentParameters={'Wavelength': 1.54, 'Coordinates': 1}
    if LayerParameters is None:
        LayerParameters={'d': 0.0, 'sigma': 0.0, 'n': 0.0+0.0j}
    if StackParameters is None:
        StackParameters={}
    if SampleParameters is None:
        SampleParameters={}
    if SimulationFunctions is None:
        SimulationFunctions={'Specular': lambda x: x, 'OffSpecular': lambda x: x}
    factory=ModelFactory()
    factory.set_layer_parameters(LayerParameters)
    factory.set_stack_parameters(StackParameters)
    factory.set_sample_parameters(SampleParameters)
    factory.set_simulation_functions(SimulationFunctions)
    factory.set_instrument_parameters(InstrumentParameters)
    Layer=factory.get_layer_class()
    Stack=factory.get_stack_class()
    Sample=factory.get_sample_class()
    Instrument=factory.get_instrument_class()
    return Instrument, Layer, Stack, Sample

if __name__=='__main__':
    InstrumentParameters={'probe': 'x-ray', 'wavelength': 1.54, 'coords': 'tth',
                          'I0': 1.0, 'res': 0.001,
                          'restype': 'no conv', 'respoints': 5, 'resintrange': 2, 'beamw': 0.01,
                          'footype': 'no corr', 'samplelen': 10.0, 'incangle': 0.0, 'pol': 'uu',
                          'Ibkg': 0.0, 'tthoff': 0.0}

    LayerParameters={'sigma': 0.0, 'dens': 1.0, 'd': 0.0, 'f': (1.0+1.0j)*1e-20,
                     'b': 0.0+0.0J, 'xs_ai': 0.0, 'magn': 0.0, 'magn_ang': 0.0}
    StackParameters={'Layers': [], 'Repetitions': 1}
    SampleParameters={'Stacks': [], 'Ambient': None, 'Substrate': None}

    (Instrument, Layer, Stack, Sample)=MakeClasses(InstrumentParameters,
                                                   LayerParameters, StackParameters, SampleParameters,
                                                   {'Specualr': lambda x: x},
                                                   'test')
    inst=Instrument(footype='gauss beam', probe='x-ray', beamw=0.04, resintrange=2, pol='uu', wavelength=1.54,
                    respoints=5, Ibkg=0.0, I0=2, samplelen=10.0, restype='fast conv', coords='2θ', res=0.001,
                    incangle=0.0)

    # BEGIN Sample DO NOT CHANGE
    Amb=Layer(b=0, d=0.0, f=(1e-20+1e-20j), dens=1.0, magn_ang=0.0, sigma=0.0, xs_ai=0.0, magn=0.0)
    topPt=Layer(b=0, d=11.0, f=58, dens=4/3.92**3, magn_ang=0.0, sigma=3.0, xs_ai=0.0, magn=0.0)
    TopFe=Layer(b=0, d=11.0, f=26, dens=2/2.866**3, magn_ang=0.0, sigma=2.0, xs_ai=0.0, magn=0.0)
    Pt=Layer(b=0, d=11.0, f=58, dens=4/3.92**3, magn_ang=0.0, sigma=2.0, xs_ai=0.0, magn=0.0)
    Fe=Layer(b=0, d=11, f=26, dens=2/2.866**3, magn_ang=0.0, sigma=2.0, xs_ai=0.0, magn=0.0)
    bufPt=Layer(b=0, d=45, f=58, dens=4/3.92**3, magn_ang=0.0, sigma=2, xs_ai=0.0, magn=0.0)
    bufFe=Layer(b=0, d=2, f=26, dens=2/2.866**3, magn_ang=0.0, sigma=2, xs_ai=0.0, magn=0.0)
    Sub=Layer(b=0, d=0.0, f=12+16, dens=2/4.2**3, magn_ang=0.0, sigma=4.0, xs_ai=0.0, magn=0.0)

    top=Stack(Layers=[TopFe, topPt], Repetitions=1)
    ML=Stack(Layers=[Fe, Pt], Repetitions=19)
    buffer=Stack(Layers=[bufFe, bufPt], Repetitions=1)

    sample=Sample(Stacks=[buffer, ML, top], Ambient=Amb, Substrate=Sub)

    # Test case for ReflFunction class
    from scipy import interpolate

    def create_dispersion_func(name='Fe'):
        path='../databases/f1f2_nist/'
        e, f1, f2=np.loadtxt(path+'%s.nff'%name.lower(), skiprows=1,
                             unpack=True)
        f1interp=interpolate.interp1d(e, f1, kind='linear')
        f2interp=interpolate.interp1d(e, f2, kind='linear')

        def f(energy):
            return f1interp(energy)-1.0J*f2interp(energy)

        return f

    fFe=ReflFunction(create_dispersion_func('Fe'), (1000,), {}, id='f(E)')
    fCo=ReflFunction(create_dispersion_func('Co'), (1000,), {}, id='f(E)')
    iprint(fFe.validate())
    iprint('Cast to array tests:')
    ltest=[fFe, 3.0, 3.0]
    iprint('Test single value: ', cast_to_array(ltest, 1000))
    iprint('Test array: ', cast_to_array(ltest, np.arange(1000, 1010)))
    iprint('Cast to array finished.')



'''
Module used to setup the default GUI logging and messaging system.
The system contains on a python logging based approach with logfile,
console output and GUI output dependent on startup options and
message logLevel.
'''

import sys
import atexit
import logging
import inspect
from numpy import seterr, seterrcall
from io import StringIO
from ..version import __version__ as str_version

# default options used if nothing is set in the configuration
CONSOLE_LEVEL, FILE_LEVEL, GUI_LEVEL=logging.WARNING, logging.DEBUG, logging.INFO

# set log levels according to options
if 'pdb' in list(sys.modules.keys()) or 'pydevd' in list(sys.modules.keys()):
    # if common debugger modules have been loaded, assume a debug run
    CONSOLE_LEVEL, FILE_LEVEL, GUI_LEVEL=logging.INFO, logging.DEBUG, logging.INFO
elif '--debug' in sys.argv:
    sys.argv.remove('--debug')
    CONSOLE_LEVEL, FILE_LEVEL, GUI_LEVEL=logging.DEBUG, logging.DEBUG, logging.INFO

def genx_exit_message():
    logging.info('*** GenX %s Logging ended ***'%str_version)


# noinspection PyUnusedLocal
def iprint(*objects, sep=None, end=None, file=None, flush=False):
    """
    A logging function that behaves like print but uses logging.info.
    """
    if sep is None:
        sep=' '
    if end is None:
        end=''#'\n'
    logging.info(sep.join(map(str, objects))+end)

class NumpyLogger(logging.getLoggerClass()):
    '''
      A logger that makes sure the actual function definition filename, lineno and function name
      is used for logging numpy floating point errors, not the numpy_logger function.
    '''

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
        curframe=inspect.currentframe()
        calframes=inspect.getouterframes(curframe, 2)
        # stack starts with:
        # (this method, debug call, debug call root logger, numpy_logger, actual function, ...)
        ignore, fname, lineno, func, ignore, ignore=calframes[4]
        # noinspection PyUnresolvedReferences
        return logging.getLoggerClass().makeRecord(self, name, level, fn, lno,
                                                   msg, args, exc_info, func=func, extra=extra)

nplogger=None

def setup_system():
    logger=logging.getLogger()
    logger.setLevel(min(CONSOLE_LEVEL, GUI_LEVEL))

    # no console logger for windows (win32gui)
    console=logging.StreamHandler(sys.__stdout__)
    formatter=logging.Formatter('%(levelname) 7s: %(message)s')
    console.setFormatter(formatter)
    console.setLevel(CONSOLE_LEVEL)
    logger.addHandler(console)

    logging.getLogger('matplotlib').setLevel(logging.WARNING)
    if min(CONSOLE_LEVEL, GUI_LEVEL)>logging.DEBUG:
        logging.getLogger('numba').setLevel(logging.WARNING)
    logging.info('*** GenX %s Logging started ***'%str_version)

    # define numpy warning behavior
    global nplogger
    old_class=logging.getLoggerClass()
    logging.setLoggerClass(NumpyLogger)
    nplogger=logging.getLogger('numpy')
    nplogger.setLevel(logging.DEBUG)
    null_handler=logging.StreamHandler(StringIO())
    null_handler.setLevel(logging.CRITICAL)
    nplogger.addHandler(null_handler)
    logging.setLoggerClass(old_class)
    seterr(divide='call', over='call', under='ignore', invalid='call')
    logging.captureWarnings(True)

    def numpy_logger(err, _flag):
        nplogger.debug('numpy floating point error encountered (%s)'%err)

    seterrcall(numpy_logger)

    # write information on program exit
    atexit.register(genx_exit_message)

def activate_logging(logfile):
    logger=logging.getLogger()
    logfile=logging.FileHandler(logfile, 'w')
    logger.setLevel(min(logger.getEffectiveLevel(), FILE_LEVEL))
    formatter=logging.Formatter('[%(levelname)s] - %(asctime)s - %(filename)s:%(lineno)i:%(funcName)s %(message)s', '')
    logfile.setFormatter(formatter)
    logfile.setLevel(FILE_LEVEL)
    logger.addHandler(logfile)
    logger.info('*** GenX %s Logging started to file ***'%str_version)

_prev_excepthook=None
def excepthook_overwrite(*exc_info):
    # making sure all exceptions are displayed on GUI and logged
    # noinspection PyTypeChecker
    logging.critical('uncought python error', exc_info=exc_info)
    # noinspection PyCallingNonCallable
    return _prev_excepthook(*exc_info)

def activate_excepthook():
    logging.debug("replacing sys.excepthook with user function")
    global _prev_excepthook
    _prev_excepthook=sys.excepthook
    sys.excepthook=excepthook_overwrite


'''
===========
SimpleLayer
===========

A plugin to allow a quick and simple creation of layers from materials created
from a crystallographic data file (.cif) available at e.g. ICSD
or the chemical formula and mass-density or crystal structure.
Materials are stored for later use.

To use the materials select any layer in the Sample tab and one material in the
Materials tab and than click the blue arrow button. If you select a Stack instead
a new layer will be created with the given material. The plugin will insert
the x-ray and neutron values ( *f* , *b* ) as well as the density and rename the layer
with the given formula.

Written by Artur Glavic
Last Changes 10/11/16
'''

import os
import wx
from wx.lib.mixins.listctrl import ListCtrlAutoWidthMixin
from math import cos, pi, sqrt
from ...gui import images as img
from .. import add_on_framework as framework
from .help_modules.materials_db import mdb, Formula, MASS_DENSITY_CONVERSION

mg=None
pymysql=None

class PluginInterface:
    def __init__(self, plugin):
        self._plugin=plugin

class RefPluginInterface(PluginInterface):
    def Update(self):
        try:
            self._plugin.sample_widget.UpdateListbox()
        except AttributeError:
            self._plugin.sample_widget.Update()

    def get_selected_layer(self):
        layer_idx=self._plugin.sample_widget.listbox.GetSelection()
        active_layer=self._plugin.sampleh.getItem(layer_idx)
        if active_layer.__class__.__name__=="Stack":
            # create a new layer to return
            self._plugin.sampleh.insertItem(layer_idx, 'Layer', 'WillChange')
            active_layer=self._plugin.sampleh.getItem(layer_idx+1)
        return active_layer

    def set_layer_name(self, name):
        layer_idx=self._plugin.sample_widget.listbox.GetSelection()
        if self._plugin.sampleh.names[layer_idx] in ['Amb', 'Sub']:
            return
        active_layer=self._plugin.sampleh.getItem(layer_idx)
        if active_layer.__class__.__name__=="Stack":
            # create a new layer to return
            layer_idx+=1
        tmpname=name
        i=1
        while tmpname in self._plugin.sampleh.names:
            tmpname='%s_%i'%(name, i)
            i+=1
        self._plugin.sampleh.names[layer_idx]=tmpname

    def material_apply(self, index, panel=None):
        formula, density=mdb[index]
        try:
            layer=self.get_selected_layer()
        except:
            dlg=wx.MessageDialog(panel,
                                 'You have to select a layer or stack before applying material',
                                 caption='Information',
                                 style=wx.OK | wx.ICON_INFORMATION)
            dlg.ShowModal()
            dlg.Destroy()
            return
        if layer:
            layer.f=formula.f()
            layer.b=formula.b()
            layer.dens=density
        name=''
        for element, count in formula:
            # get rid of isotope unusable characters
            element=element.replace('{', '').replace('}', '').replace('^', 'i')
            if count==1:
                name+="%s"%element
            elif float(count)==int(count):
                name+="%s%i"%(element, count)
            else:
                name+=("%s%s"%(element, count)).replace('.', '_')
        self.set_layer_name(name)
        self.Update()

class SimplePluginInterface(PluginInterface):
    def Update(self):
        self._plugin.Update()

    def get_selected_layer(self):
        row=self._plugin.sample_widget.grid.GetGridCursorRow()
        layer_type=self._plugin.sample_widget.sample_table.GetValue(row, 1)
        return row, layer_type

    def material_apply(self, index, panel=None):
        formula, density=mdb[index]
        row, layer_type=self.get_selected_layer()
        if layer_type is None:
            dlg=wx.MessageDialog(panel,
                                 'You have to select a layer to apply the material to',
                                 caption='Information',
                                 style=wx.OK | wx.ICON_INFORMATION)
            dlg.ShowModal()
            dlg.Destroy()
            return
        g=self._plugin.sample_widget.grid
        t=self._plugin.sample_widget.sample_table
        if layer_type=='Mixure':
            col=g.GetGridCursorCol()
            if not col in [2, 4]:
                dlg=wx.MessageDialog(panel,
                                     'Select SLD cell to apply material to',
                                     caption='Information',
                                     style=wx.OK | wx.ICON_INFORMATION)
                dlg.ShowModal()
                dlg.Destroy()
                return
            if self._plugin.sample_widget.inst_params['probe']=='x-ray':
                SLD=mdb.SLDx(index).real
            else:
                SLD=mdb.SLDn(index).real
            t.SetValue(row, col, str(SLD))
            t.GetView().ForceRefresh()
        else:
            t.SetValue(row, 2, formula.estr())
            t.GetView().ForceRefresh()
            t.SetValue(row, 4, '%g'%mdb.dens_mass(index))

class Plugin(framework.Template):
    _refplugin=None

    @property
    def refplugin(self):
        # check if reflectivity plugin is None or destoryed, try to connect
        if not self._refplugin:
            self._init_refplugin()
        return self._refplugin

    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        self.parent=parent
        # on the right side, add a list of materials with their density to selct from
        materials_panel=self.NewDataFolder('Materials')
        materials_sizer=wx.BoxSizer(wx.HORIZONTAL)
        self.materials_panel=wx.Panel(materials_panel)
        self.create_materials_list()
        materials_sizer.Add(self.materials_panel, 1, wx.EXPAND | wx.GROW | wx.ALL)
        materials_panel.SetSizer(materials_sizer)
        materials_panel.Layout()

    def _init_refplugin(self):
        ph=self.parent.plugin_control.plugin_handler
        if 'Reflectivity' in ph.loaded_plugins:
            # connect to the reflectivity plugin for layer creation
            self._refplugin=RefPluginInterface(
                ph.loaded_plugins['Reflectivity'])
        elif 'SimpleReflectivity' in ph.loaded_plugins:
            # connect to the reflectivity plugin for layer creation
            self._refplugin=SimplePluginInterface(
                ph.loaded_plugins['SimpleReflectivity'])
        else:
            dlg=wx.MessageDialog(self.materials_panel, 'Reflectivity plugin must be loaded',
                                 caption='Information',
                                 style=wx.OK | wx.ICON_WARNING)
            dlg.ShowModal()
            dlg.Destroy()
            self._refplugin=None

    def create_materials_list(self):
        '''
          Create a list of materials and it's graphical representation as
          well as a toolbar.
        '''
        self.known_materials=mdb
        # self.tool_panel=wx.Panel(self.materials_panel)
        self.materials_list=MaterialsList(self.materials_panel, self.known_materials)
        self.sizer_vert=wx.BoxSizer(wx.VERTICAL)
        self.materials_panel.SetSizer(self.sizer_vert)

        self.create_toolbar()

        # self.sizer_vert.Add(self.tool_panel, proportion=0, flag=wx.EXPAND, border=5)
        self.sizer_vert.Add((-1, 2))
        self.sizer_vert.Add(self.materials_list, proportion=1, flag=wx.EXPAND, border=5)
        # self.tool_panel.SetSizer(self.sizer_hor)

    def create_toolbar(self):
        self.toolbar=wx.ToolBar(self.materials_panel, style=wx.TB_FLAT | wx.TB_HORIZONTAL)

        dpi_scale_factor=wx.GetApp().dpi_scale_factor
        tb_bmp_size=int(dpi_scale_factor*20)

        newid=wx.NewId()
        self.toolbar.AddTool(newid, label='Add',
                             bitmap=wx.Bitmap(img.add.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Add a material to the list')
        self.materials_panel.Bind(wx.EVT_TOOL, self.material_add, id=newid)

        newid=wx.NewId()
        self.toolbar.AddTool(newid, label='Delete',
                             bitmap=wx.Bitmap(img.delete.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='Delete selected materials')
        self.materials_panel.Bind(wx.EVT_TOOL, self.material_delete, id=newid)

        newid=wx.NewId()
        self.toolbar.AddTool(newid, label='Apply',
                             bitmap=wx.Bitmap(img.start_fit.GetImage().Scale(tb_bmp_size, tb_bmp_size)),
                             shortHelp='New Layer/Apply to Layer')
        self.materials_panel.Bind(wx.EVT_TOOL, self.material_apply, id=newid)

        # self.sizer_hor=wx.BoxSizer(wx.HORIZONTAL)
        self.sizer_vert.Add(self.toolbar, proportion=0, flag=wx.EXPAND, border=2)
        self.toolbar.Realize()

    def material_add(self, event):
        dialog=MaterialDialog(self.parent)
        if dialog.ShowModal()==wx.ID_OK:
            self.materials_list.AddItem(dialog.GetResult())
        dialog.Destroy()

    def material_delete(self, event):
        self.materials_list.DeleteItem()

    def material_apply(self, event):
        index=self.materials_list.GetFirstSelected()
        self.refplugin.material_apply(index, panel=self.materials_panel)

class MaterialsList(wx.ListCtrl, ListCtrlAutoWidthMixin):
    '''
    The ListCtrl for the materials data.
    '''

    def __init__(self, parent, materials_list):
        wx.ListCtrl.__init__(self, parent, -1,
                             style=wx.LC_REPORT | wx.LC_VIRTUAL)
        ListCtrlAutoWidthMixin.__init__(self)
        self.materials_list=materials_list
        self.parent=parent

        font=self.GetFont()
        font.SetPointSize(9)
        self.SetFont(font)

        # Set list length
        self.SetItemCount(len(materials_list))

        # Set the column headers
        for col, (text, width) in enumerate([
            ("Chemical Formula", 80),
            ("n [10⁻⁶Å⁻²]", 60),
            ("kα [rₑ/Å⁻³]", 60),
            ("FU/Å³", 60),
            ("g/cm³", 60)
            ]):
            self.InsertColumn(col, text, width=width)

        self.setResizeColumn(0)

    def OnSelectionChanged(self, evt):
        if not self.toggleshow:
            indices=self._GetSelectedItems()
            indices.sort()
            if not indices==self.show_indices:
                self.data_cont.show_data(indices)
                self._UpdateData('Show data set flag toggled',
                                 data_changed=True)
                # Forces update of list control
                self.SetItemCount(self.data_cont.get_count())
        evt.Skip()

    def OnGetItemText(self, item, col):
        if col==4:
            return "%.3f"%self.materials_list.dens_mass(item)
        elif col==3:
            return "%.4f"%self.materials_list.dens_FU(item).real
        elif col==2:
            return "%.3f"%self.materials_list.SLDx(item).real
        elif col==1:
            return "%.3f"%self.materials_list.SLDn(item).real
        else:
            prev_names=[str(mi[0]) for mi in self.materials_list[:item]]
            name=str(self.materials_list[item][0])
            prev_count=prev_names.count(name)
            if prev_count>0:
                name+='-%i'%(prev_count+1)
            return name

    def _GetSelectedItems(self):
        ''' _GetSelectedItems(self) --> indices [list of integers]
        Function that yields a list of the currently selected items
        position in the list. In order of selction, i.e. no order.
        '''
        indices=[self.GetFirstSelected()]
        while indices[-1]!=-1:
            indices.append(self.GetNextSelected(indices[-1]))

        # Remove the last will be -1
        indices.pop(-1)
        return indices

    def _CheckSelected(self, indices):
        '''_CheckSelected(self, indices) --> bool
        Checks so at least data sets are selcted, otherwise show a dialog box
        and return False
        '''
        # Check so that one dataset is selected
        if len(indices)==0:
            dlg=wx.MessageDialog(self, 'At least one data set has to be selected', caption='Information',
                                 style=wx.OK | wx.ICON_INFORMATION)
            dlg.ShowModal()
            dlg.Destroy()
            return False
        return True

    def DeleteItem(self):
        index=self.GetFirstSelected()
        item=self.materials_list[index]
        item_formula=''
        for element, count in item[0]:
            if count==1:
                item_formula+="%s"%element
            elif float(count)==int(count):
                item_formula+="%s%i"%(element, count)
            else:
                item_formula+="%s(%f)"%(element, count)

        # Create the dialog box
        dlg=wx.MessageDialog(self, 'Remove material %s?'%item_formula,
                             caption='Remove?', style=wx.YES_NO | wx.ICON_QUESTION)

        # Show the dialog box
        if dlg.ShowModal()==wx.ID_YES:
            self.materials_list.pop(index)
            # Update the list
            self.SetItemCount(len(self.materials_list))
            self.RefreshItems(index, len(self.materials_list))

        dlg.Destroy()

    def AddItem(self, item):
        index=0
        while index<len(self.materials_list) and self.materials_list[index][0]<item[0]:
            index+=1
        self.materials_list.insert(index, item)
        self.SetItemCount(len(self.materials_list))
        self.RefreshItems(index, len(self.materials_list))

class MaterialDialog(wx.Dialog):
    """
      Dialog to get material information from chemical formula and atomic density.
      Atomic density can ither be entered manually, by using unit cell parameter,
      by massdensity or by loading a .cif crystallographical file.
    """

    def __init__(self, parent):
        wx.Dialog.__init__(self, parent, name='New Material')
        self.extracted_elements=Formula([])
        self._create_entries()

    def _create_entries(self):
        base_layout=wx.BoxSizer(wx.VERTICAL)

        table=wx.GridBagSizer(5, 2)

        self.formula_entry=wx.TextCtrl(self, size=(100, 25))
        self.formula_entry.Bind(wx.EVT_TEXT, self.OnFormulaChanged)
        table.Add(wx.StaticText(self, label="Formula:"), (0, 0), flag=wx.ALIGN_CENTER)
        table.Add(self.formula_entry, (0, 1), span=(1, 2), flag=wx.EXPAND)

        table.Add(wx.StaticText(self, label="Extracted Elements:"), (1, 0),
                  span=(1, 3), flag=wx.ALIGN_CENTER)
        self.formula_display=wx.TextCtrl(self, size=(150, 100),
                                         style=wx.TE_MULTILINE | wx.TE_READONLY)
        #        self.formula_display.Enable(False)
        table.Add(self.formula_display, (2, 0), span=(1, 3), flag=wx.EXPAND)

        table.Add(wx.StaticText(self, label="1. Unit Cell Parameters:"), (3, 0),
                  span=(1, 3), flag=wx.ALIGN_CENTER)
        table.Add(wx.StaticText(self, label="a [Å]"), (4, 0), flag=wx.ALIGN_CENTER)
        self.a_entry=wx.TextCtrl(self, size=(50, 25))
        table.Add(self.a_entry, (5, 0), flag=wx.EXPAND)
        table.Add(wx.StaticText(self, label="b [Å]"), (4, 1), flag=wx.ALIGN_CENTER)
        self.b_entry=wx.TextCtrl(self, size=(50, 25))
        table.Add(self.b_entry, (5, 1), flag=wx.EXPAND)
        table.Add(wx.StaticText(self, label="c [Å]"), (4, 2), flag=wx.ALIGN_CENTER)
        self.c_entry=wx.TextCtrl(self, size=(50, 25))
        table.Add(self.c_entry, (5, 2), flag=wx.EXPAND)

        table.Add(wx.StaticText(self, label="α"), (6, 0), flag=wx.ALIGN_CENTER)
        self.alpha_entry=wx.TextCtrl(self, size=(50, 25))
        self.alpha_entry.SetValue('90')
        table.Add(self.alpha_entry, (7, 0), flag=wx.EXPAND)
        table.Add(wx.StaticText(self, label="β"), (6, 1), flag=wx.ALIGN_CENTER)
        self.beta_entry=wx.TextCtrl(self, size=(50, 25))
        self.beta_entry.SetValue('90')
        table.Add(self.beta_entry, (7, 1), flag=wx.EXPAND)
        table.Add(wx.StaticText(self, label="γ"), (6, 2), flag=wx.ALIGN_CENTER)
        self.gamma_entry=wx.TextCtrl(self, size=(50, 25))
        self.gamma_entry.SetValue('90')
        table.Add(self.gamma_entry, (7, 2), flag=wx.EXPAND)

        table.Add(wx.StaticText(self, label="FUs"), (8, 0), flag=wx.ALIGN_CENTER)
        self.FUs_entry=wx.TextCtrl(self, size=(50, 25))
        self.FUs_entry.SetValue('1')
        table.Add(self.FUs_entry, (9, 0), flag=wx.ALIGN_CENTER)
        cif_button=wx.Button(self, label="From .cif File...")
        cif_button.Bind(wx.EVT_BUTTON, self.OnLoadCif)
        table.Add(cif_button, (8, 1), span=(2, 2), flag=wx.ALIGN_CENTER)

        global mg, pymysql
        if mg is None:
            try:
                global MPRester
                from pymatgen.ext.matproj import MPRester
                import pymatgen as mg
            except ImportError:
                pass
        if mg is None:
            mg_txt=wx.StaticText(self, label="Install PyMatGen for Materials Project Query")
            table.Add(mg_txt, (10, 0), span=(1, 3), flag=wx.ALIGN_CENTER)
        else:
            mg_button=wx.Button(self, label="Query The Materials Project")
            mg_button.Bind(wx.EVT_BUTTON, self.OnQuery)
            table.Add(mg_button, (10, 0), span=(1, 3), flag=wx.ALIGN_CENTER)

        if pymysql is None:
            try:
                import pymysql
            except ImportError:
                pass
        if pymysql is None:
            cod_txt=wx.StaticText(self, label="Install pymysql for COD Query")
            table.Add(cod_txt, (11, 0), span=(1, 3), flag=wx.ALIGN_CENTER)
        else:
            cod_button=wx.Button(self, label="Query Crystallography Open DB")
            cod_button.Bind(wx.EVT_BUTTON, self.OnCODQuery)
            table.Add(cod_button, (11, 0), span=(1, 3), flag=wx.ALIGN_CENTER)

        for entry in [self.a_entry, self.b_entry, self.c_entry,
                      self.alpha_entry, self.beta_entry, self.gamma_entry,
                      self.FUs_entry]:
            entry.Bind(wx.EVT_TEXT, self.OnUnitCellChanged)

        table.Add(wx.StaticText(self, label="2. Physical Parameter:"), (12, 0),
                  span=(1, 3), flag=wx.ALIGN_CENTER)
        self.mass_density=wx.TextCtrl(self, size=(70, 25))
        self.mass_density.Bind(wx.EVT_TEXT, self.OnMassDensityChange)
        table.Add(wx.StaticText(self, label="Mass Density [g/cm³]:"), (13, 0))
        table.Add(self.mass_density, (13, 1), span=(1, 1))

        table.Add(wx.StaticText(self, label="Result from 1. or 2.:"), (14, 0), span=(1, 3))
        self.result_density=wx.TextCtrl(self, size=(150, 25))
        table.Add(wx.StaticText(self, label="Density [FU/Å³]:"), (15, 0))
        table.Add(self.result_density, (15, 1), span=(1, 2))

        buttons=self.CreateButtonSizer(wx.OK | wx.CANCEL)

        base_layout.Add(table, 1, wx.ALIGN_CENTER | wx.TOP)
        base_layout.Add(buttons, 0, wx.ALIGN_RIGHT)
        self.SetSizerAndFit(base_layout)

    def OnFormulaChanged(self, event):
        text=self.formula_entry.GetValue()
        for ign_char in [" ", "\t", "_", "-"]:
            text=text.replace(ign_char, "")
        if text=="":
            self.extracted_elements=Formula([])
            self.formula_display.SetValue('')
            return
        formula=Formula.from_str(text)
        self.extracted_elements=formula
        self.formula_display.SetValue(formula.describe())
        self.OnMassDensityChange(None)

    def OnUnitCellChanged(self, event):
        params=[]
        for entry in [self.a_entry, self.b_entry, self.c_entry,
                      self.alpha_entry, self.beta_entry, self.gamma_entry,
                      self.FUs_entry]:
            try:
                params.append(float(entry.GetValue()))
            except ValueError:
                return
        if params[3]==90 and params[4]==90 and params[5]==90:
            if params[0]==params[1] and params[0]==params[2]:
                self.density='%s/(%g**3)'%(params[6], params[0])
            else:
                self.density='%s/(%g*%g*%g)'%(params[6], params[0], params[1], params[2])
        else:
            # calculate general unit cell volume (triclinic formula applicable to all structures)
            alpha=params[3]*pi/180.
            beta=params[4]*pi/180.
            gamma=params[5]*pi/180.
            V=params[0]*params[1]*params[2]
            V*=sqrt(1-cos(alpha)**2-cos(beta)**2-cos(gamma)**2+2*cos(alpha)*cos(beta)*cos(gamma))
            self.density='%s/%g'%(params[6], V)
        self.result_density.SetValue(self.density)

    def OnMassDensityChange(self, event):
        fu_mass=self.extracted_elements.mFU()
        try:
            mass_density=float(self.mass_density.GetValue())
        except ValueError:
            return
        self.density="%g*%g/%g"%(mass_density, MASS_DENSITY_CONVERSION, fu_mass)
        self.result_density.SetValue(self.density)

    def OnLoadCif(self, event):
        fd=wx.FileDialog(self,
                         message="Open a (.cif) File...",
                         wildcard='Crystallographic Information File|*.cif;*.CIF|All Files|*',
                         defaultFile='crystal.cif',
                         style=wx.FD_OPEN | wx.FD_CHANGE_DIR)
        if fd.ShowModal()==wx.ID_OK:
            filename=fd.GetPath()
            self.extract_cif(filename)
        fd.Destroy()

    def OnQuery(self, event):
        key='NdHi2bTnJ9WDS1sU'
        a=MPRester(key)
        formula=''
        for element, number in self.extracted_elements:
            if element.startswith('^'):
                element=element.split('}')[-1]
            formula+='%s%g'%(element, number)
        res=a.get_data(formula)
        if type(res) is not list:
            return
        if len(res)>1:
            # more then one structure available, ask for user input to select appropriate
            items=[]
            for i, ri in enumerate(res):
                cs=ri['spacegroup']['crystal_system']
                sgs=ri['spacegroup']['symbol']
                frm=ri['full_formula']
                v=ri['volume']
                dens=ri['density']
                items.append(
                    '%i: %s (%s) | UC Formula: %s\n     Density: %s g/cm³ | UC Volume: %s'%
                    (i+1, sgs, cs, frm, dens, v))
                if ri['tags'] is not None:
                    items[-1]+='\n     '+';'.join(ri['tags'][:3])
            dia=wx.SingleChoiceDialog(self,
                                      'Several entries have been found, please select appropriate:',
                                      'Select correct database entry',
                                      items)
            if not dia.ShowModal()==wx.ID_OK:
                return None
            res=res[dia.GetSelection()]
        else:
            res=res[0]
        return self.analyze_cif(res['cif'])

    def OnCODQuery(self, event):
        db=pymysql.connect(host='sql.crystallography.net',
                           user='cod_reader',
                           password=None,
                           database='cod')
        c=db.cursor()
        formula=[]
        for element, number in self.extracted_elements:
            if element.startswith('^'):
                element=element.split('}')[-1]
            if number!=1:
                formula.append('%s%g'%(element, number))
            else:
                formula.append('%s'%element)
        formula=' '.join(sorted(formula))
        c.execute('select a,b,c,alpha,beta,gamma,'
                  'Z,vol,sg,chemname,mineral,commonname,'
                  'authors,title,journal,year,'
                  'formula,calcformula'
                  ' from data where formula like "- %s -"'
                  ' and method != "theoretical"'
                  ' and status is NULL order by file'%formula)
        res=c.fetchall()
        if len(res)>0:
            # more then one structure available, ask for user input to select appropriate
            items=[]
            for i, ri in enumerate(res):
                a, b, c=ri[:3]
                if ri[6] is None:
                    try:
                        f, calcf=ri[16:18]
                        fbase=f.split()[1]
                        cbase=calcf.split()[1]
                        i=0
                        while i<len(fbase) and not fbase[i].isdigit():
                            i+=1
                        if i==len(fbase):
                            fbase+='1'
                        if i==len(cbase):
                            cbase+='1'
                        Z=float(cbase[i:])/float(fbase[i:])
                    except Exception:
                        Z=1.
                else:
                    Z=ri[6]
                V=ri[7] or a*b*c
                sgs=ri[8]
                dens=self.extracted_elements.mFU()*Z/V/MASS_DENSITY_CONVERSION
                name=ri[10] or ri[11] or ri[9] or ""
                text='%s\n%s\n%s (%s)'%tuple(ri[12:16])
                items.append(
                    '%i: %s (%s) |  Density: %.3g g/cm³ | UC Volume: %s\n%s'%
                    (i+1, name, sgs, dens, V, text))
            dia=wx.SingleChoiceDialog(self,
                                      'Several entries have been found, please select appropriate:',
                                      'Select correct database entry',
                                      items)
            if not dia.ShowModal()==wx.ID_OK:
                return None
            res=res[dia.GetSelection()]
        else:
            return None
        for value, entry in zip(res,
                                [self.a_entry,
                                 self.b_entry,
                                 self.c_entry,
                                 self.alpha_entry,
                                 self.beta_entry,
                                 self.gamma_entry,
                                 self.FUs_entry]):
            entry.SetValue(str(value or ""))

    def GetResult(self):
        return self.extracted_elements, self.result_density.GetValue()

    def extract_cif(self, filename):
        """
          Try to get unit cell and formula unit information from a .cif file.
        """
        if not os.path.exists(filename):
            return
        txt=open(filename).read()
        return self.analyze_cif(txt)

    def analyze_cif(self, txt):
        cell_params=[1., 1., 1., 90., 90., 90., 1.]
        composition=''
        file_lines=txt.splitlines()
        for line in file_lines:
            if line.startswith('_cell_length_a'):
                cell_params[0]=round(float(line.split()[1].split('(')[0]), 3)
            if line.startswith('_cell_length_b'):
                cell_params[1]=round(float(line.split()[1].split('(')[0]), 3)
            if line.startswith('_cell_length_c'):
                cell_params[2]=round(float(line.split()[1].split('(')[0]), 3)
            if line.startswith('_cell_angle_alpha'):
                cell_params[3]=round(float(line.split()[1].split('(')[0]), 3)
            if line.startswith('_cell_angle_beta'):
                cell_params[4]=round(float(line.split()[1].split('(')[0]), 3)
            if line.startswith('_cell_angle_gamma'):
                cell_params[5]=round(float(line.split()[1].split('(')[0]), 3)
            if line.startswith('_cell_formula_units_Z'):
                cell_params[6]=int(float(line.split()[1]))
            if line.startswith('_chemical_formula_structural'):
                composition=line.strip().split(None, 1)[1].replace(")", "").replace("(", "").replace("'", "").replace(
                    '"', '')
            if line.startswith('_chemical_formula_sum') and composition=='':
                composition=line.strip().split(None, 1)[1].replace("'", "").replace('"', '')
        self.formula_entry.SetValue(composition)
        self.OnFormulaChanged(None)
        for value, entry in zip(cell_params,
                                [self.a_entry,
                                 self.b_entry,
                                 self.c_entry,
                                 self.alpha_entry,
                                 self.beta_entry,
                                 self.gamma_entry,
                                 self.FUs_entry]):
            entry.SetValue(str(value))
        self.OnUnitCellChanged(None)



from numpy import *
from numba import cuda
import numba
import math, cmath

@cuda.jit(numba.void(numba.float64[:], numba.float64, numba.complex128[:],
                     numba.float64[:], numba.float64[:], numba.float64[:]))
def ReflNB(theta, lamda, n, d, sigma, Rout):
    # Thread id in a 1D block
    tx=cuda.threadIdx.x
    # Block id in a 1D grid
    ty=cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw=cuda.blockDim.x
    # Compute flattened index inside the array
    ai=tx+ty*bw
    if ai>=theta.shape[0]:
        return

    layers=d.shape[0]
    n0=n[-1]

    k=2.*pi/lamda

    pre1=2.*n0*k
    pre2=(pi/180.)

    Qi=pre1*cmath.sqrt((n[0]/n0)**2-math.cos(theta[ai]*pre2)**2)

    Qj=pre1*cmath.sqrt((n[1]/n0)**2-math.cos(theta[ai]*pre2)**2)
    # Fresnel reflectivity for the interfaces
    rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[0]**2)
    Aj=rpj
    Qi=Qj

    for lj in range(2, layers):
        Qj=pre1*cmath.sqrt((n[lj]/n0)**2-math.cos(theta[ai]*pre2)**2)
        # Fresnel reflectivity for the interfaces
        rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[lj-1]**2)

        pj=cmath.exp(1.0j*d[lj-1]*Qi)
        part=Aj*pj
        Aj=(rpj+part)/(1.+part*rpj)

        Qi=Qj
    Rout[ai]=abs(Aj)**2

@cuda.jit(numba.void(numba.float64[:], numba.float64, numba.complex128[:],
                     numba.float64[:], numba.float64[:], numba.complex128[:]))
def AmpNB(theta, lamda, n, d, sigma, Aout):
    # Thread id in a 1D block
    tx=cuda.threadIdx.x
    # Block id in a 1D grid
    ty=cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw=cuda.blockDim.x
    # Compute flattened index inside the array
    ai=tx+ty*bw
    if ai>=theta.shape[0]:
        return

    layers=d.shape[0]
    n0=n[-1]

    k=2.*pi/lamda

    pre1=2.*n0*k
    pre2=(pi/180.)

    Qi=pre1*cmath.sqrt((n[0]/n0)**2-math.cos(theta[ai]*pre2)**2)

    Qj=pre1*cmath.sqrt((n[1]/n0)**2-math.cos(theta[ai]*pre2)**2)
    # Fresnel reflectivity for the interfaces
    rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[0]**2)
    Aj=rpj
    Qi=Qj

    for lj in range(2, layers):
        Qj=pre1*cmath.sqrt((n[lj]/n0)**2-math.cos(theta[ai]*pre2)**2)
        # Fresnel reflectivity for the interfaces
        rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[lj-1]**2)

        pj=cmath.exp(1.0j*d[lj-1]*Qi)
        part=Aj*pj
        Aj=(rpj+part)/(1.+part*rpj)

        Qi=Qj
    Aout[ai]=Aj

def Refl(theta, lamda, n, d, sigma, return_int=True):
    threadsperblock=32
    blockspergrid=(theta.shape[0]+(threadsperblock-1))//threadsperblock

    Ctheta=cuda.to_device(theta)
    Cn=cuda.to_device(n.astype(complex128))
    Cd=cuda.to_device(d)
    Csigma=cuda.to_device(sigma)

    if return_int:
        CRout=cuda.device_array(shape=theta.shape, dtype=float64)
        ReflNB[blockspergrid, threadsperblock](Ctheta, lamda, Cn, Cd, Csigma, CRout)
        return CRout.copy_to_host()
    else:
        CAout=cuda.device_array(shape=theta.shape, dtype=complex128)
        AmpNB[blockspergrid, threadsperblock](Ctheta, lamda, Cn, Cd, Csigma, CAout)
        return CAout.copy_to_host()

@cuda.jit(numba.void(numba.float64[:], numba.float64, numba.complex128[:],
                     numba.float64[:], numba.float64[:], numba.float64[:]))
def ReflQNB(Q, Q0, n, d, sigma, Rout):
    # Thread id in a 1D block
    tx=cuda.threadIdx.x
    # Block id in a 1D grid
    ty=cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw=cuda.blockDim.x
    # Compute flattened index inside the array
    qi=tx+ty*bw
    if qi>=Q.shape[0]:
        return

    layers=d.shape[0]
    n0=n[-1]

    Qi=cmath.sqrt((n[0]**2-n0**2)*Q0**2+n0**2*Q[qi]**2)

    Qj=cmath.sqrt((n[1]**2-n0**2)*Q0**2+n0**2*Q[qi]**2)
    # Fresnel reflectivity for the interfaces
    rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[0]**2)
    Aj=rpj
    Qi=Qj

    for lj in range(2, layers):
        Qj=cmath.sqrt((n[lj]**2-n0**2)*Q0**2+n0**2*Q[qi]**2)
        # Fresnel reflectivity for the interfaces
        rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[lj-1]**2)

        pj=cmath.exp(1.0j*d[lj-1]*Qi)
        part=Aj*pj
        Aj=(rpj+part)/(1.+part*rpj)

        Qi=Qj
    Rout[qi]=abs(Aj)**2

@cuda.jit(numba.void(numba.float64[:], numba.float64, numba.complex128[:],
                     numba.float64[:], numba.float64[:], numba.complex128[:]))
def AmpQNB(Q, Q0, n, d, sigma, Aout):
    # Thread id in a 1D block
    tx=cuda.threadIdx.x
    # Block id in a 1D grid
    ty=cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw=cuda.blockDim.x
    # Compute flattened index inside the array
    qi=tx+ty*bw
    if qi>=Q.shape[0]:
        return

    layers=d.shape[0]
    points=Q.shape[0]
    n0=n[-1]

    Qi=cmath.sqrt((n[0]**2-n0**2)*Q0**2+n0**2*Q[qi]**2)

    Qj=cmath.sqrt((n[1]**2-n0**2)*Q0**2+n0**2*Q[qi]**2)
    # Fresnel reflectivity for the interfaces
    rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[0]**2)
    Aj=rpj
    Qi=Qj

    for lj in range(2, layers):
        Qj=cmath.sqrt((n[lj]**2-n0**2)*Q0**2+n0**2*Q[qi]**2)
        # Fresnel reflectivity for the interfaces
        rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[lj-1]**2)

        pj=cmath.exp(1.0j*d[lj-1]*Qi)
        part=Aj*pj
        Aj=(rpj+part)/(1.+part*rpj)

        Qi=Qj
    Aout[qi]=Aj

def ReflQ(Q, lamda, n, d, sigma, return_int=True):
    threadsperblock=32
    blockspergrid=(Q.shape[0]+(threadsperblock-1))//threadsperblock

    Q0=4.*pi/lamda

    CQ=cuda.to_device(Q)
    Cn=cuda.to_device(n.astype(complex128))
    Cd=cuda.to_device(d)
    Csigma=cuda.to_device(sigma)

    if return_int:
        CRout=cuda.device_array(shape=Q.shape, dtype=float64)
        ReflQNB[blockspergrid, threadsperblock](CQ, Q0, Cn, Cd, Csigma, CRout)
        return CRout.copy_to_host()
    else:
        CAout=cuda.device_array(shape=Q.shape, dtype=complex128)
        AmpQNB[blockspergrid, threadsperblock](CQ, Q0, Cn, Cd, Csigma, CAout)
        return CAout.copy_to_host()

@cuda.jit(numba.void(numba.float64[:], numba.float64[:], numba.complex128[:, :],
                     numba.float64[:], numba.float64[:], numba.float64[:]))
def Refl_nvary2NB(theta, lamda, n, d, sigma, Rout):
    # Thread id in a 1D block
    tx=cuda.threadIdx.x
    # Block id in a 1D grid
    ty=cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw=cuda.blockDim.x
    # Compute flattened index inside the array
    ai=tx+ty*bw
    if ai>=theta.shape[0]:
        return

    layers=d.shape[0]

    pre2=(pi/180.)

    n0=n[-1, ai]
    ki=2.*pi/lamda[ai]
    pre1=2.*n0*ki

    Qi=pre1*cmath.sqrt((n[0, ai]/n0)**2-math.cos(theta[ai]*pre2)**2)

    Qj=pre1*cmath.sqrt((n[1, ai]/n0)**2-math.cos(theta[ai]*pre2)**2)
    # Fresnel reflectivity for the interfaces
    rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[0]**2)
    Aj=rpj
    Qi=Qj

    for lj in range(2, layers):
        Qj=pre1*cmath.sqrt((n[lj, ai]/n0)**2-math.cos(theta[ai]*pre2)**2)
        # Fresnel reflectivity for the interfaces
        rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[lj-1]**2)

        pj=cmath.exp(1.0j*d[lj-1]*Qi)
        part=Aj*pj
        Aj=(rpj+part)/(1.+part*rpj)

        Qi=Qj
    Rout[ai]=abs(Aj)**2

@cuda.jit(numba.void(numba.float64[:], numba.float64[:], numba.complex128[:, :],
                     numba.float64[:], numba.float64[:], numba.complex128[:]))
def Amp_nvary2NB(theta, lamda, n, d, sigma, Aout):
    # Thread id in a 1D block
    tx=cuda.threadIdx.x
    # Block id in a 1D grid
    ty=cuda.blockIdx.x
    # Block width, i.e. number of threads per block
    bw=cuda.blockDim.x
    # Compute flattened index inside the array
    ai=tx+ty*bw
    if ai>=theta.shape[0]:
        return

    layers=d.shape[0]

    pre2=(pi/180.)

    n0=n[-1, ai]
    ki=2.*pi/lamda[ai]
    pre1=2.*n0*ki

    Qi=pre1*cmath.sqrt((n[0, ai]/n0)**2-math.cos(theta[ai]*pre2)**2)

    Qj=pre1*cmath.sqrt((n[1, ai]/n0)**2-math.cos(theta[ai]*pre2)**2)
    # Fresnel reflectivity for the interfaces
    rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[0]**2)
    Aj=rpj
    Qi=Qj

    for lj in range(2, layers):
        Qj=pre1*cmath.sqrt((n[lj, ai]/n0)**2-math.cos(theta[ai]*pre2)**2)
        # Fresnel reflectivity for the interfaces
        rpj=(Qj-Qi)/(Qj+Qi)*cmath.exp(-Qj*Qi/2.*sigma[lj-1]**2)

        pj=cmath.exp(1.0j*d[lj-1]*Qi)
        part=Aj*pj
        Aj=(rpj+part)/(1.+part*rpj)

        Qi=Qj
    Aout[ai]=Aj

def Refl_nvary2(theta, lamda, n, d, sigma, return_int=True):
    threadsperblock=32
    blockspergrid=(theta.shape[0]+(threadsperblock-1))//threadsperblock

    Ctheta=cuda.to_device(theta)
    Clamda=cuda.to_device(lamda)
    Cn=cuda.to_device(n.astype(complex128))
    Cd=cuda.to_device(d)
    Csigma=cuda.to_device(sigma)

    if return_int:
        CRout=cuda.device_array(shape=theta.shape, dtype=float64)
        Refl_nvary2NB[blockspergrid, threadsperblock](Ctheta, Clamda, Cn, Cd, Csigma, CRout)
        return CRout.copy_to_host()
    else:
        CAout=cuda.device_array(shape=theta.shape, dtype=complex128)
        Amp_nvary2NB[blockspergrid, threadsperblock](Ctheta, Clamda, Cn, Cd, Csigma, CAout)
        return CAout.copy_to_host()



'''<h1>UserFuncs</h1>
A plugin so that users can include their own user function into
the model script and execute them inside the GUI of GenX.<p>

The following design criteria exists:
The function has to be defined inside the script.
The function has to take zero input arguments.
<p>
Thus the following example can serve as an template:
<pre>def myuserfunc():
    # Do something
    print 'It works!'
</pre>
This should be added somewhere in the script. This provides a great way
to, for example, dump some internal data into a file or for checking the status
of some variables.
'''
from .. import add_on_framework as framework
import types, wx, io, traceback

class Plugin(framework.Template):
    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        self.menu=self.NewMenu('User funcs')
        self.parent=parent
        self.StatusMessage('Sucessfully loaded UserFuncs...')

    def OnSimulate(self, event):
        '''OnSimulate(self, event) --> None
        
        Updates stuff after simulation
        '''
        model=self.GetModel()
        # locate all functions in the model.script_module
        funcs=[eval('model.script_module.%s'%name, globals(), {'model': model}) \
               for name in dir(model.script_module) \
               if type(eval('model.script_module.%s'%name, globals(), {'model': model})
                       )==types.FunctionType]
        # Find the functions that are defined in the script and takes zero
        # input parameters
        user_funcs=[f.__name__ for f in funcs if \
                    f.__module__=='genx_script_module' \
                    and f.__code__.co_argcount==0]
        # Remove all the previous functions
        self.clear_menu()
        # Lets add our user functions to our custo menu
        for name in user_funcs:
            menuitem=wx.MenuItem(self.menu, wx.NewId(), name,
                                 'Evaluate %s'%name, wx.ITEM_NORMAL)
            self.menu.AppendItem(menuitem)
            self.parent.Bind(wx.EVT_MENU, self.eh_menu_choice, menuitem)

    def clear_menu(self):
        '''clear_menu(self) --> None
        
        Clears the menu from all items present in it
        '''
        [self.menu.RemoveItem(item) for item in self.menu.GetMenuItems()]

    def eh_menu_choice(self, event):
        '''eh_menu_choice(self, event)
        
        event handler for the choice of an menuitem in the User Functions 
        menu. Executes the function as defined in the script. 
        With error catching.
        '''
        menu=event.GetEventObject()
        menuitem=menu.FindItemById(event.Id)
        fname=menuitem.GetLabel()
        model=self.GetModel()
        # Now try to evaluate the function
        self.StatusMessage('Trying to evaluate %s'%fname)
        try:
            exec('%s()'%fname, model.script_module.__dict__)
        except Exception as e:
            # abit of traceback
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            tb_string=outp.getvalue()
            outp.close()
            self.ShowWarningDialog('Error in evaluating the' \
                                   ' function: %s.\n The error gave the following traceback:' \
                                   '\n%s'%(fname, tb_string))
            self.StatusMessage('error in function %s'%fname)
        else:
            self.StatusMessage('%s sucessfully evaluated'%fname)



'''Library to calculate the reflectivity with the formalism presented in
S. A. Stepanov and S. K. Sinha, Phys. Rev. B 61, 15302.

All the matrices/arrays in this module has the following index order:
index: Meaning
1: matrix row
2: matrix column
3: Layer index
4: scanning angle (g0, incident angle)

'''

from .math_utils import *
from .int_lay_xmean import calc_xrmr_Xmean
from functools import reduce
from genx.core.custom_logging import iprint

# mpy_limit = 1e-9 corresponds to an angle of 89.998 deg.
def calc_refl(g_0, lamda, chi0, A, B, C, M, d, mag_limit=1e-8, mpy_limit=1e-9):
    ''' Calculate the reflectivity according to the
    recursion matrix formalism as given by
    S. A. Stepanov and S. K. Sinha, Phys. Rev. B 61, 15302.
    https://doi.org/10.1103/PhysRevB.61.15302
    Note that this formalism has the first item in the parameters
    describing the layers (chi0, A, B, C, d) as the vacuum. 
    Any ambient layer is ignored!
    g_0 = sin Phi_0 where Phi_0 is the incident angle
    lamda - the wavelength of the experiment
    chi0 - the non-resonant suceptibility of the layers
    A - the resonant non-magnetic term of the suceptibility
    B - the resonant magnetic term linear to the magnetization (circulair dichroism)
    C - the resonant magnetic term quadratic to the magnetization.
    M - the UNIT vector of the magnetization vector direction
    d - the thickness of the layers
    mag_limit - the limit where B and C are assumed to be zero and the 
    material are assumed to be non-magnetic
    mpy_limit - magnet parallel to y limit the angle to the y-axis where the sample is assumed
    to be parallel to the y-direction

    Sizes of input arrays:
    g_0: # data points
    lamda: Either a scalar or a array with size # data points
    d: An array of size # layers
    A, B, C, chi0: either all arrays with size # layers or 2 D arrays of size (# layers, # data points)
    M: either an array of size (3, # layers) or (3, # layers, # data points)
    The if the case of more dimensions is used for one array it has to applied to all the other arrays.

    Note that the first layer should be the ambient and the last the substrate!
    '''
    chi, non_mag, mpy=create_chi(g_0, lamda, chi0, A, B, C, M, d,
                                 mag_limit=mag_limit, mpy_limit=mpy_limit)

    return do_calc(g_0, lamda, chi, d, non_mag, mpy)

def calc_refl_int_lay(g_0, lamda, chi0, A, B, C, M, d, sigma, sigma_l, sigma_u, dd_l, dd_u,
                      dmag_l, dmag_u,
                      mag_limit=1e-8, mpy_limit=1e-9):
    '''Calculate the reflectivity according to the recursion matrix formalism as calc_refl.
    This function incorporates a interface layer for the magnetism of thicknesses dd_l and dd_u.
    the interfaces has rouhnesses sigma_l, sigma_u and sigma

    Note that the first layer should be the ambient and the last the substrate!
    '''
    chi_b, non_mag_b, mpy_b=create_chi(g_0, lamda, chi0, A, B, C, M, d,
                                       mag_limit=mag_limit, mpy_limit=mpy_limit)
    scale=(1.+dmag_l)
    if len(B.shape)==2:
        scale=scale[:, np.newaxis]*np.ones(g_0.shape)
    chi_l, non_mag_l, mpy_l=create_chi(g_0, lamda, chi0, A, B*scale, C*scale**2, M, dd_l,
                                       mag_limit=mag_limit, mpy_limit=mpy_limit)
    scale=(1.+dmag_u)
    if len(B.shape)==2:
        scale=scale[:, np.newaxis]*np.ones(g_0.shape)
    chi_u, non_mag_u, mpy_u=create_chi(g_0, lamda, chi0, A, B*scale, C*scale**2, M, dd_u,
                                       mag_limit=mag_limit, mpy_limit=mpy_limit)

    d=(d-dd_l-dd_u)

    # Setting up g0 to be the last index
    g_0=g_0*np.ones(chi_b[0][0].shape[0], dtype=np.complex128)[:, np.newaxis]

    u_b, S_b=calc_u_S(chi_b, g_0, mpy_b, non_mag_b)
    u_l, S_l=calc_u_S(chi_l, g_0, mpy_l, non_mag_l)
    u_u, S_u=calc_u_S(chi_u, g_0, mpy_u, non_mag_u)

    X_lu=dot4(inv4(S_l[:, :, :-1, :]), S_u[:, :, 1:, :])
    X_l=dot4(inv4(S_b[:, :, :-1, :]), S_l[:, :, :-1, :])
    X_u=dot4(inv4(S_u[:, :, 1:, :]), S_b[:, :, 1:, :])

    d=d[:, np.newaxis]*np.ones(g_0.shape[1], dtype=np.complex128)
    dd_u=dd_u[:, np.newaxis]*np.ones(g_0.shape[1], dtype=np.complex128)
    dd_l=dd_l[:, np.newaxis]*np.ones(g_0.shape[1], dtype=np.complex128)
    sigma=sigma[:, np.newaxis]*np.ones(g_0.shape[1], dtype=np.complex128)
    sigma_l=sigma_l[:, np.newaxis]*np.ones(g_0.shape[1], dtype=np.complex128)
    sigma_u=sigma_u[:, np.newaxis]*np.ones(g_0.shape[1], dtype=np.complex128)

    Xmean=calc_xrmr_Xmean(lamda, X_l, X_lu, X_u, u_b, u_l, u_u, dd_u, dd_l, sigma, sigma_l, sigma_u)

    W_tot=recursion_matrix(Xmean, chi_b, d, g_0, lamda, u_b)
    return W_tot[2]

def create_chi(g_0, lamda, chi0, A, B, C, M, d, mag_limit=1e-8, mpy_limit=1e-9):
    A=A.astype(np.complex128)
    B=B.astype(np.complex128)
    C=C.astype(np.complex128)

    m_x=M[:, 0];
    m_y=M[:, 1];
    m_z=M[:, 2]
    # print 'my: ', m_y, m_x, m_z
    chi_xx=(chi0+A+C*m_x*m_x)
    chi_yy=(chi0+A+C*m_y*m_y)
    chi_zz=(chi0+A+C*m_z*m_z)
    chi_xy=(-1.0J*B*m_z+C*m_x*m_y)
    chi_yx=(1.0J*B*m_z+C*m_x*m_y)
    chi_xz=(1.0J*B*m_y+C*m_x*m_z)
    chi_zx=(-1.0J*B*m_y+C*m_x*m_z)
    chi_yz=(-1.0J*B*m_x+C*m_y*m_z)
    chi_zy=(1.0J*B*m_x+C*m_y*m_z)
    chi=((chi_xx, chi_xy, chi_xz), (chi_yx, chi_yy, chi_yz), (chi_zx, chi_zy, chi_zz))

    # Take into account non-magnetic materials:
    non_mag=np.bitwise_and(np.abs(B)<mag_limit, np.abs(C)<mag_limit)
    # Ignore the ambient (vacuum)
    non_mag[0]=False

    # Take into account the matrix singularity arising when M||Y
    mpy=np.bitwise_and(1.0-np.abs(m_y)<mpy_limit, np.bitwise_not(non_mag))

    return chi, non_mag, mpy

def calc_u_S(chi, g_0, mpy, non_mag):
    '''Calculates the interface matrix S and the u values.
    '''
    chi_xx, chi_xy, chi_xz=chi[0]
    chi_yx, chi_yy, chi_yz=chi[1]
    chi_zx, chi_zy, chi_zz=chi[2]
    # Check if the chi_xx has the correct format already or not
    if len(chi_xx.shape)<2:
        trans=np.ones(g_0.shape, dtype=np.complex128)
        chi_xx=trans*chi_xx[:, np.newaxis]
        chi_xy=trans*chi_xy[:, np.newaxis]
        chi_xz=trans*chi_xz[:, np.newaxis]
        chi_yx=trans*chi_yx[:, np.newaxis]
        chi_yy=trans*chi_yy[:, np.newaxis]
        chi_yz=trans*chi_yz[:, np.newaxis]
        chi_zx=trans*chi_zx[:, np.newaxis]
        chi_zy=trans*chi_zy[:, np.newaxis]
        chi_zz=trans*chi_zz[:, np.newaxis]

    n_x=np.sqrt(1.0-g_0**2)
    q1=1+chi_zz
    q2=n_x*(chi_xz+chi_zx)
    q3=(chi_xz*chi_zx+chi_yz*chi_zy-(1+chi_zz)*(g_0**2+chi_yy)
        -(1+chi_xx)*(g_0**2+chi_zz))
    q4=n_x*(chi_xy*chi_yz+chi_yx*chi_zy-
            (chi_xz+chi_zx)*(g_0**2+chi_yy))
    q5=((1+chi_xx)*((g_0**2+chi_yy)*(g_0**2+chi_zz)-
                    chi_yz*chi_zy)
        -chi_xy*chi_yx*(g_0**2+chi_zz)
        -chi_xz*chi_zx*(g_0**2+chi_yy)
        +chi_xy*chi_zx*chi_yz+chi_yx*chi_xz*chi_zy)
    # Note assuming C = 0 => q2 = 0 and q4 = 0
    # And this yields the following eigenstates
    # c = np.sqrt(q3**2 - 4*q1*q5)
    # u1 = np.sqrt((-q3 - c)/2.0/q1)
    # u2 = np.sqrt((-q3 + c)/2.0/q1)
    # u3 = -u1
    # u4 = -u2
    # End simplifaction
    # Proper solution to a 4'th degree polynomial
    u1, u2, u3, u4=roots4thdegree(q1, q2, q3, q4, q5)
    # print u1[:,ind];print u2[:,ind];print u3[:,ind];print u4[:,ind]
    # Special case M||X for finding errors in the code:
    # u1 = -np.sqrt(g_0**2 + trans*(chi0 + A - B)[:, np.newaxis])
    # u2 = np.sqrt(g_0**2 + trans*(chi0 + A + B)[:, np.newaxis])
    # u3 = -np.sqrt(g_0**2 + trans*(chi0 + A + B)[:, np.newaxis])
    # u4 = np.sqrt(g_0**2 + trans*(chi0 + A - B)[:, np.newaxis])
    # End special case
    # I am lazy and simply sort the roots in ascending order to get the
    # right direction of the light later
    u_temp=np.array((u1, u2, u3, u4))
    pos=np.argsort(u_temp.imag, axis=0)
    u_temp=np.choose(pos, u_temp)
    u=np.zeros(u_temp.shape, dtype=np.complex128)
    u[0]=u_temp[3]
    u[1]=u_temp[2]
    u[2]=u_temp[1]
    u[3]=u_temp[0]

    D=((chi_xz+u*n_x)*(chi_zx+u*n_x)-
       (1.0-u**2+chi_xx)*(g_0**2+chi_zz))
    # These expressions can throw a runtime warning if D has an incorrect value
    # The different special cases are handled later by non-magnetic flags as well
    # as m//y flag.
    old_error_vals=np.seterr(invalid='ignore')
    P_x=(chi_xy*(g_0**2+chi_zz)-
         chi_zy*(chi_xz+u*n_x))/D
    P_z=(chi_zy*(1.0-u**2+chi_xx)-
         chi_xy*(chi_zx+u*n_x))/D
    np.seterr(invalid=old_error_vals['invalid'])

    S=np.zeros((4, 4, chi_xx.shape[0], g_0.shape[1]), dtype=np.complex128)
    # The ambient layer
    S[0, 0, 0]=1.0
    S[0, 2, 0]=1.0
    S[1, 1, 0]=1.0
    S[1, 3, 0]=1.0
    S[2, 0, 0]=g_0[0]
    S[2, 2, 0]=-g_0[0]
    S[3, 1, 0]=g_0[0]
    S[3, 3, 0]=-g_0[0]
    u[0, 0]=g_0[0]
    u[1, 0]=g_0[0]
    u[2, 0]=-g_0[0]
    u[3, 0]=-g_0[0]
    # The rest of the multilayers
    v=(u*P_x-n_x*P_z)[:, 1:]
    w=P_x[:, 1:]
    S[0, :, 1:]=1.0
    S[1, :, 1:]=v
    S[2, :, 1:]=u[:, 1:]
    S[3, :, 1:]=w

    # Handle the layers that are non-magnetic
    # Removing to test if roughness calcs are affected
    chi=chi_xx[non_mag]
    nm_u1=np.sqrt(g_0[non_mag]**2+chi)
    nm_u2=-nm_u1
    sqr_eps=np.sqrt(1+chi)
    S[0, 0, non_mag]=1.0
    S[0, 1, non_mag]=0.0
    S[0, 2, non_mag]=1.0
    S[0, 3, non_mag]=0.0
    S[1, 0, non_mag]=0.0
    S[1, 1, non_mag]=sqr_eps
    S[1, 2, non_mag]=0.0
    S[1, 3, non_mag]=sqr_eps
    S[2, 0, non_mag]=nm_u1
    S[2, 1, non_mag]=0.0
    S[2, 2, non_mag]=nm_u2
    S[2, 3, non_mag]=0.0
    S[3, 0, non_mag]=0.0
    S[3, 1, non_mag]=nm_u1/sqr_eps
    S[3, 2, non_mag]=0.0
    S[3, 3, non_mag]=nm_u2/sqr_eps
    u[0, non_mag]=nm_u1
    u[1, non_mag]=nm_u1
    u[2, non_mag]=nm_u2
    u[3, non_mag]=nm_u2

    # Take into account the matrix singularity arising when M||Y
    if np.any(mpy):
        # print 'M||Y calcs activated'
        delta=chi_xz[mpy]**2*(1+chi_xx[mpy])
        nx=n_x[mpy]
        mpy_u1=np.sqrt(g_0[mpy]**2+chi_yy[mpy])
        mpy_u3=-mpy_u1
        mpy_u2=np.sqrt(g_0[mpy]**2+chi_zz[mpy])
        mpy_u4=-mpy_u2
        S[0, 0, mpy]=1.0
        S[0, 1, mpy]=0.0
        S[0, 2, mpy]=1.0
        S[0, 3, mpy]=0.0
        S[1, 0, mpy]=0.0
        S[1, 1, mpy]=-(mpy_u2*chi_xz[mpy]+nx*(1+chi_xx[mpy]))/(nx**2-delta)
        S[1, 2, mpy]=0.0
        S[1, 3, mpy]=-(mpy_u4*chi_xz[mpy]+nx*(1+chi_xx[mpy]))/(nx**2-delta)
        S[2, 0, mpy]=mpy_u1
        S[2, 1, mpy]=0.0
        S[2, 2, mpy]=mpy_u3
        S[2, 3, mpy]=0.0
        S[3, 0, mpy]=0.0
        S[3, 1, mpy]=-(mpy_u2*nx+chi_xz[mpy])/(nx**2-delta)
        S[3, 2, mpy]=0.0
        S[3, 3, mpy]=-(mpy_u4*nx+chi_xz[mpy])/(nx**2-delta)
        u[0, mpy]=mpy_u1
        u[1, mpy]=mpy_u2
        u[2, mpy]=mpy_u3
        u[3, mpy]=mpy_u4

    return u, S

def recursion_matrix(X, chi, d, g_0, lamda, u):
    ''' Use the recursion matrix algorithm to form the W matrix for a multilayer '''
    kappa=2*np.pi/lamda
    Fp=np.zeros((2, 2, chi[0][0].shape[0], g_0.shape[1]), dtype=np.complex128)
    Fm=np.zeros((2, 2, chi[0][0].shape[0], g_0.shape[1]), dtype=np.complex128)
    Fp[0, 0]=np.exp(-1.0J*u[0]*kappa*d)
    Fp[1, 1]=np.exp(-1.0J*u[1]*kappa*d)
    Fm[0, 0]=np.exp(-1.0J*u[2]*kappa*d)
    Fm[1, 1]=np.exp(-1.0J*u[3]*kappa*d)
    Fp=Fp[:, :, 1:]
    Fm=Fm[:, :, 1:]

    Xtt=X[:2, :2]
    Xtr=X[:2, 2:]
    Xrt=X[2:, :2]
    Xrr=X[2:, 2:]
    Mtt=dot2(inv2(Fp), inv2(Xtt))
    Mtr=-dot2(Mtt, dot2(Xtr, Fm))
    Mrt=dot2(Xrt, inv2(Xtt))
    Mrr=dot2(Xrr-dot2(Mrt, Xtr), Fm)

    def reduce_func(W, i):
        return calc_W(W[0], W[1], W[2], W[3], Mtt[:, :, i, :], Mtr[:, :, i, :], Mrt[:, :, i, :], Mrr[:, :, i, :])

    W0tt=np.zeros((2, 2, g_0.shape[1]), dtype=np.complex128)
    W0tt[0, 0]=1.0
    W0tt[1, 1]=1.0
    W0rr=np.zeros((2, 2, g_0.shape[1]), dtype=np.complex128)
    W0rr[0, 0]=1.0
    W0rr[1, 1]=1.0
    W0rt=np.zeros((2, 2, g_0.shape[1]), dtype=np.complex128)
    W0tr=np.zeros((2, 2, g_0.shape[1]), dtype=np.complex128)

    W=reduce(reduce_func, list(range(Mtt.shape[2])), (W0tt, W0tr, W0rt, W0rr))
    return W

def do_calc(g_0, lamda, chi, d, non_mag, mpy):
    ''' Do the calculation to form the mulatilayer reflectivity matrix, W'''
    # Setting up g0 to be the last index
    g_0=g_0*np.ones(chi[0][0].shape[0], dtype=np.complex128)[:, np.newaxis]
    u, S=calc_u_S(chi, g_0, mpy, non_mag)
    X=dot4(inv4(S[:, :, :-1, :]), S[:, :, 1:, :])
    d=d[:, np.newaxis]*np.ones(g_0.shape[1], dtype=np.complex128)
    W_tot=recursion_matrix(X, chi, d, g_0, lamda, u)
    return W_tot[2]

def calc_W(Wtt, Wtr, Wrt, Wrr, Mtt, Mtr, Mrt, Mrr):
    ''' Calculate the Wn+1 matrices for recursion formalism'''

    ident=np.zeros(Wtt.shape, dtype=np.complex128)
    ident[0, 0]=1.0
    ident[1, 1]=1.0

    A=dot2(Mtt, inv2(ident-dot2(Wtr, Mrt)))
    B=dot2(Wrr, inv2(ident-dot2(Mrt, Wtr)))

    Wtt_next=dot2(A, Wtt)
    Wtr_next=Mtr+dot2(dot2(A, Wtr), Mrr)
    Wrt_next=Wrt+dot2(dot2(B, Mrt), Wtt)
    Wrr_next=dot2(B, Mrr)
    return Wtt_next, Wtr_next, Wrt_next, Wrr_next

def calc_nonres(g_0, lamda, chi0, d):
    trans=np.ones(g_0.shape, dtype=np.complex128)
    chi_xx=trans*chi0[:, np.newaxis]
    epsilon=1.0+chi_xx
    kappa=2*np.pi/lamda
    d=np.ones(g_0.shape)*d[:, np.newaxis]
    g_0=g_0*np.ones(chi0.shape)[:, np.newaxis]
    u1=np.sqrt(g_0**2+chi0[:, np.newaxis])
    u2=-u1

    S=np.zeros((4, 4, chi0.shape[0], g_0.shape[1]), dtype=np.complex128)
    S[0, 0, 0]=1;
    S[0, 2, 0]=1;
    S[1, 1, 0]=1;
    S[1, 3, 0]=1
    S[2, 0, 0]=g_0[0];
    S[2, 2, 0]=-g_0[0];
    S[3, 1, 0]=g_0[0];
    S[3, 3, 0]=-g_0[0]
    # S_v = np.array([[1, 0, 1, 0], [0, 1, 0, 1],
    #                [g_0, 0, -g_0, 0], [0, g_0, 0, -g_0]])
    e=np.sqrt(epsilon[1:])
    S[0, 0, 1:]=1;
    S[0, 2, 1:]=1
    S[1, 1, 1:]=e;
    S[1, 3, 1:]=e
    S[2, 0, 1:]=u1[1:];
    S[2, 2, 1:]=u2[1:]
    S[3, 1, 1:]=u1[1:]/e;
    S[3, 3, 1:]=u2[1:]/e
    X=dot4(inv4(S[:, :, :-1, :]), S[:, :, 1:, :])
    # print 'X: ', X

    Fp=np.zeros((2, 2, chi0.shape[0], g_0.shape[1]), dtype=np.complex128)
    Fm=np.zeros((2, 2, chi0.shape[0], g_0.shape[1]), dtype=np.complex128)
    Fp[0, 0]=np.exp(-1.0J*u1*kappa*d);
    Fp[1, 1]=np.exp(-1.0J*u1*kappa*d)
    Fm[0, 0]=np.exp(-1.0J*u2*kappa*d);
    Fm[1, 1]=np.exp(-1.0J*u2*kappa*d)
    Fp=Fp[:, :, 1:, :];
    Fm=Fm[:, :, 1:, :]
    # print 'F: ', Fp
    Xtt=X[:2, :2]
    Xtr=X[:2, 2:]
    Xrt=X[2:, :2]
    Xrr=X[2:, 2:]
    # print Xrr.shape, Xtr.shape, Fm.shape
    Mtt=dot2(inv2(Fp), inv2(Xtt))
    # print det2(Xtt), det2(Xtr), det2(Xtr), det2(Xrr)
    Mtr=-dot2(Mtt, dot2(Xtr, Fm))
    Mrt=dot2(Xrt, inv2(Xtt))
    # print Mrt
    Mrr=dot2(Xrr-dot2(Mrt, Xtr), Fm)

    return Mrt

if __name__=='__main__':
    import time

    # alpha = np.arange(0.01, 10.0, 0.01)
    alpha=np.arange(0.25, 30.0, 0.025)
    g_0=np.sin(alpha*np.pi/180.0)

    F10=16.0+10.0J
    F11=12.0+6.0J
    F1m1=20.0+14.0J
    fco=(27.0-0.001J)
    fpt=78.0-0.001J
    n_a_co=2./2.86**3
    n_a_pt=4./3.92**3
    l=15.77
    l_array=l*np.ones(g_0.shape)
    # l = 1.07
    re=2.8179402894e-5
    chi0_co=-l**2*re/np.pi*n_a_co*fco
    chi0_pt=-l**2*re/np.pi*n_a_pt*fpt

    N=100
    # Single layer BEGIN
    A=np.array([0, ]+[l**2*re/np.pi*n_a_co*(F11+F1m1), ]*N+[0, ],
               dtype=np.complex128)  # [:, np.newaxis]*np.ones(g_0.shape)
    B=np.array([0, ]+[l**2*re/np.pi*n_a_co*(F11-F1m1), ]*N+[0.0],
               dtype=np.complex128)  # [:, np.newaxis]*np.ones(g_0.shape)
    C=np.array([0, ]+[l**2*re/np.pi*n_a_co*(2*F10-F11-F1m1), ]*N+[0],
               dtype=np.complex128)  # [:, np.newaxis]*np.ones(g_0.shape)

    chi0=np.array([0.0, ]+[chi0_co, ]*N+[chi0_pt, ])  # [:,np.newaxis]*np.ones(g_0.shape)
    d=np.array([1.0, ]+[100.0, ]*N+[10.0])
    M=np.array([[1.0, 0.0, 0.0]]+[[1.0, 0.0, 0.0]]*N+[[1.0, 0.0, 0.0]])  # [:,:,np.newaxis]*np.ones(g_0.shape)
    # SINGLE LAYER END

    # Single interface BEGIN
    # A = np.array([1e-8, l**2*re/np.pi*n_a_co*(F11 + F1m1)], dtype = np.complex128)
    # B = np.array([1e-8,l**2*re/np.pi*n_a_co*(F11 - F1m1)], dtype = np.complex128)
    # C = np.array([0,l**2*re/np.pi*n_a_co*(2*F10 - F11 - F1m1)*0], dtype = np.complex128)
    #
    # chi0 = np.array([0.0, chi0_co])
    # d = np.array([0.0, 100.0])
    # theta = 0.0*np.pi/180.0
    # phi = 0.0*np.pi/180.0
    # M = np.array([[1.0, 0.0, 0.0], [np.cos(theta)*np.cos(phi), np.sin(theta)*np.cos(phi), np.sin(phi)]])
    # SINGLE interface END

    ##N = 1
    ##A = np.array([1e-9] + [1e-9, l**2*re/np.pi*n_a_pt*(F11 + F1m1)/5., 1e-9, l**2*re/np.pi*n_a_pt*(F11 + F1m1)/5.]*N + [1e-9], dtype = np.complex128)
    ##B = np.array([1e-9] + [1e-9, l**2*re/np.pi*n_a_pt*(F11 - F1m1)/5., 1e-9, l**2*re/np.pi*n_a_pt*(F11 - F1m1)/5.]*N + [1e-9], dtype = np.complex128)

    ##chi0 = np.array([0.0] + [chi0_pt, chi0_pt, chi0_co, chi0_pt]*N + [chi0_pt])
    ##d = np.array([0.0] + [9.0, 5.0, 10.0, 5.0]*N + [0.0])
    ##M = np.array([[1.0, 0.0, 0.0]]  + [[1.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 0.0, 0.0], [1.0, 0.0, 0.0]]*N + [[1.0, 0.0, 0.0]])

    t1=time.time()
    W=calc_refl(g_0, l_array, chi0, A, B, C, M, d)
    t2=time.time()
    W=calc_refl_int_lay(g_0, l_array, chi0, A, B, C, M, d, d*0, d*0, d*0, d*0, d*0, d*0, d*0)
    t3=time.time()
    iprint("No roughness, simulation time: ", t2-t1, "s")
    iprint("Roughness factors, simulation time ", t3-t2, "s")

    # W = calc_nonres(g_0, l, chi0, d)
    Ias=2*(W[0, 0]*W[0, 1].conj()+W[1, 0]*W[1, 1].conj()).imag/(
                np.abs(W[0, 0])**2+np.abs(W[1, 0])**2+np.abs(W[0, 1])**2+np.abs(W[1, 1])**2)
    Itot=(np.abs(W[0, 0])**2+np.abs(W[1, 0])**2+np.abs(W[0, 1])**2+np.abs(W[1, 1])**2)/2
    I=np.abs(W)**2
    trans=np.ones(W.shape, dtype=np.complex128);
    trans[0, 1]=1.0J;
    trans[1, 1]=-1.0J;
    trans=trans/np.sqrt(2)
    Wc=dot2(trans, dot2(W, np.conj(inv2(trans))))
    Ic=np.abs(Wc)**2
    # I2 = np.abs(W2)**2
    ref=np.loadtxt('MAG09235.dat')
    from pylab import *

    figure()
    subplot(211)
    title("Verification of calculations (full lines)\n against Stephanov's x-ray server(symbols)")
    semilogy(alpha, Itot, 'r', alpha, (np.abs(Wc[0, 0])**2+np.abs(Wc[1, 0])**2), 'b', alpha,
             (np.abs(Wc[0, 1])**2+np.abs(Wc[1, 1])**2), 'g')
    semilogy(ref[:, 0], (ref[:, 2]+ref[:, 3])/2., 'r.', ref[:, 0], ref[:, 2], 'b.', ref[:, 0], ref[:, 3], 'g.')
    legend(('Itot', 'circ+', 'circ-'))
    ylabel('Reflectivity [arb. u.]')
    subplot(212)
    plot(alpha, Ias, 'r')
    plot(ref[:, 0], ref[:, -1], 'r.')
    ylabel('Asymmetry')
    xlabel('Incident angle [deg]')
    # print Ias
    # show()

    # try to do some profiling
    import pstats


    calc_refl_int_lay(g_0, l_array, chi0, A, B, C, M, d, d*0, d*0, d*0, d*0, d*0, d*0, d*0)
    # cProfile.runctx("calc_refl(g_0, l_array, chi0, A, B, C, M, d)", globals(), locals(), "Profile.prof")
    # cProfile.runctx("calc_refl_int_lay(g_0, l_array, chi0, A, B, C, M, d, d*0, d*0, d*0, d*0, d*0, d*0, d*0)",
    #                globals(), locals(), "Profile.prof")

    s=pstats.Stats("Profile.prof")
    s.strip_dirs().sort_stats("time").print_stats()



'''data_loader_wx.py

Implements the GUI side of data_loader as DataLoaderController which
takes care of the use of the DataLoaders.
'''
import wx, os, io, traceback

from .utils import PluginHandler
from genx.core.config import config
from genx.core.custom_logging import iprint

head, tail=os.path.split(__file__)
# Look only after the file name and not the ending since
# the file ending can be pyc if compiled...
__FILENAME__=tail.split('.')[0]
# This assumes that plugin is under the current dir may need
# changing
__MODULE_DIR__=head
if __MODULE_DIR__!='/':
    __MODULE_DIR__+='/'

class PluginController:
    def __init__(self, parent):
        self.plugin_handler=PluginHandler(parent, __MODULE_DIR__
                                          , 'data_loaders')
        self.parent=parent
        self.plugin_handler.load_plugin('auto')

    def load_default(self):
        try:
            plugin_name=config.get('data handling', 'data loader')
            self.LoadPlugin(plugin_name)
        except Exception as S:
            iprint('Could not locate the data loader parameter or the data loader. Error:')
            iprint(S.__str__())
            iprint('Proceeding with laoding the default data loader.')
            self.LoadPlugin('default')

    def LoadPlugin(self, plugin):
        '''LoadPlugin(self, plugin) --> None

        Loads a data handler note that there is no UnLoad function
        since only one DataHandler can be plugged in at a time.
        '''
        # Unload the plugins
        names=self.plugin_handler.loaded_plugins.copy()
        try:
            [self.plugin_handler.unload_plugin(pl) for pl in names]
            self.parent.SetStatusText('Unloaded data loader %s'%list(names.keys())[0])
        except:
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            tbtext=outp.getvalue()
            outp.close()
            ShowErrorDialog(self.parent, 'Can NOT unload plugin object'+ \
                            list(names.keys())[0]+'\nPython traceback below:\n\n'+tbtext)
        try:
            self.plugin_handler.load_plugin(plugin)
            self.parent.SetStatusText('Loaded data loader: %s'%plugin)
        except:
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            tbtext=outp.getvalue()
            outp.close()
            ShowErrorDialog(self.parent, 'Can NOT load plugin '+plugin \
                            +'\nPython traceback below:\n\n'+tbtext)

    def ShowDialog(self):
        '''ShowDialog(self) --> None

        Shows a dialog boc for the user to choose a data loader.
        '''
        cur_plugin=list(self.plugin_handler.loaded_plugins.keys())[0]
        plugin_list=self.plugin_handler.get_possible_plugins()
        dlg=PluginDialog(self.parent, plugin_list, cur_plugin,
                         self.LoadPlugin)
        dlg.ShowModal()
        dlg.Destroy()

class PluginDialog(wx.Dialog):
    def __init__(self, parent, plugin_list, current_plugin,
                 load_plugin_func=None):
        wx.Dialog.__init__(self, parent, -1, 'Choose a data loader')

        self.load_plugin_func=load_plugin_func

        choice_sizer=wx.BoxSizer(wx.HORIZONTAL)
        sizer=wx.BoxSizer(wx.VERTICAL)

        choice_text=wx.StaticText(self, -1, 'Data loaders: ')
        self.choice_control=wx.Choice(self, -1, choices=plugin_list)
        self.choice_control.SetStringSelection(current_plugin)
        choice_sizer.Add(choice_text, 0, wx.ALIGN_CENTER_VERTICAL)
        choice_sizer.Add(self.choice_control, 0, wx.ALIGN_CENTER_VERTICAL)

        # Add the Dialog buttons
        button_sizer=wx.StdDialogButtonSizer()
        okay_button=wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        apply_button=wx.Button(self, wx.ID_APPLY)
        apply_button.SetDefault()
        button_sizer.AddButton(apply_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))
        button_sizer.Realize()
        # Add some eventhandlers
        self.Bind(wx.EVT_BUTTON, self.on_apply, okay_button)
        self.Bind(wx.EVT_BUTTON, self.on_apply, apply_button)

        sizer.Add((20, 20), 0)
        sizer.Add(choice_sizer, 0, wx.ALIGN_CENTRE | wx.ALL)
        line=wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.TOP, 20)
        sizer.Add(button_sizer)

        self.SetSizer(sizer)

        sizer.Fit(self)
        self.Layout()

    def on_apply(self, event):
        if self.load_plugin_func is not None:
            self.load_plugin_func(self.choice_control.GetStringSelection())
        event.Skip()

# Utility Dialog functions..
def ShowInfoDialog(frame, message):
    dlg=wx.MessageDialog(frame, message,
                         'Information',
                         wx.OK | wx.ICON_INFORMATION
                         )
    dlg.ShowModal()
    dlg.Destroy()

def ShowErrorDialog(frame, message, position=''):
    dlg=wx.MessageDialog(frame, message,
                         'ERROR',
                         wx.OK | wx.ICON_ERROR
                         )
    dlg.ShowModal()
    dlg.Destroy()

def ShowWarningDialog(frame, message):
    dlg=wx.MessageDialog(frame, message, 'Warning',
                         wx.OK | wx.ICON_ERROR
                         )
    dlg.ShowModal()
    dlg.Destroy()



'''
Module to handle plugins including a template class for writing
plugins. 
'''

import os
import wx, io, traceback
from dataclasses import dataclass

from genx.core.config import BaseConfig, Configurable
from .utils import PluginHandler

head, tail=os.path.split(__file__)
# Look only after the file name and not the ending since
# the file ending can be pyc if compiled... 
__FILENAME__=tail.split('.')[0]
# This assumes that plugin is under the current dir may need 
# changing
__MODULE_DIR__=head
if __MODULE_DIR__!='/':
    __MODULE_DIR__+='/'

class Template:
    ''' A template class for handling plugins. Note that using the 
    New* methods will automatically remove them when removing the plugin.
    Otherwise the programmer, if he/she makes more advanced changes in the gui,
    have to take care of the deletion of objects.
    '''

    # TODO: Implement an Bind/Unbind mehtods.
    # TODO: Add a Load data function Needs change in data as well.
    def __init__(self, parent):
        '''__init__(self, parent)
        This method should be overloaded.
        '''
        self.parent=parent
        self.plot_pages=[]
        self.input_pages=[]
        self.data_pages=[]
        self.menus=[]
        self.parent.input_notebook.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGED, self.OnInputPageChanged)

    def OnInputPageChanged(self, evt):
        tpage, fpage=evt.GetSelection(), evt.GetOldSelection()
        if fpage!=tpage:
            tpanel=self.parent.input_notebook.GetPage(tpage)
            if tpanel in self.input_pages:
                self.InputPageChanged(self.parent.input_notebook.GetPageText(tpage))
        evt.Skip()

    def InputPageChanged(self, pname):
        # Can be used to react to own plugin pages being selected
        pass

    def NewPlotFolder(self, name, pos=-1):
        '''NewPlotFolder(self, name) --> wx.Panel
        
        Creates a new Folder in the Plot part of the panels. Returns
        a wx.Panel which can be used to create custom controls. 
        o not forget to use RemovePlotfolder in the Destroy method.
        '''
        panel=wx.Panel(self.parent.plot_notebook, -1)
        self.parent.plot_notebook.AddPage(panel, name)
        self.plot_pages.append(panel)

        return panel

    def NewInputFolder(self, name, pos=-1):
        '''NewInputFolder(self, name, pos = -1) --> wx.Panel
        
        Creates a new Folder in the Input part of the panels. Returns
        a wx.Panel which can be used to create custom controls. 
        o not forget to use RemoveInputfolder in the Destroy method.
        '''
        panel=wx.Panel(self.parent.input_notebook, -1)
        self.parent.input_notebook.AddPage(panel, name)
        self.input_pages.append(panel)

        return panel

    def NewDataFolder(self, name, pos=-1):
        '''NewDataFolder(self, name, pos = -1) --> wx.Panel
        
        Creates a new Folder in the data part of the panels. Returns
        a wx.Panel which can be used to create custom controls. 
        o not forget to use RemoveInputfolder in the Destroy method.
        '''
        panel=wx.Panel(self.parent.data_notebook, -1)
        self.parent.data_notebook.AddPage(panel, name)
        self.data_pages.append(panel)

        return panel

    def NewMenu(self, name):
        '''NewMenu(self, name) --> wx.Menu
        
        Creates an top menu that can be used to control the plugin. Remeber
        to also implement RemoveMenu in the Destroy method.
        '''
        menu=wx.Menu()
        self.parent.main_frame_menubar.Append(menu, name)
        self.menus.append(name)

        return menu

    def StatusMessage(self, text):
        '''StatusMessage(self, text) --> None
        
        Method that sets the staustext in the main window
        '''
        self.parent.main_frame_statusbar.SetStatusText(text, 1)

    def ShowErrorDialog(self, message):
        '''ShowErrorDialog(self, message) --> None
        
        Shows an error dialog with message [string]
        '''
        ShowErrorDialog(self.parent, message)

    def ShowInfoDialog(self, message):
        '''ShowInfoDialog(self, message) --> None
        
        Shows an info dialog with message [string]
        '''
        ShowInfoDialog(self.parent, message)

    def ShowWarningDialog(self, message):
        '''ShowWarningDialog(self, message) --> None
        
        Shows an warning dialog with message [string]
        '''
        ShowWarningDialog(self.parent, message)

    def ShowQuestionDialog(self, message):
        '''ShowWarningDialog(self, message) --> None
        
        Shows an warning dialog with message [string]
        '''
        return ShowQuestionDialog(self.parent, message)

    def GetModel(self):
        '''GetModel(self) --> model 
        
        Returns the model currently in use. This is a pointer to the model
        object thus it will automatically always conatin the newest information.        
        '''
        return self.parent.model

    def GetSolverControl(self):
        '''GetSolverControl(self) --> solver_control
        
        Returns the solver_control object that controls all aspects of
        the calculational part of the fitting.
        '''
        return self.parent.solver_control

    def SetModelScript(self, script):
        '''SetModelScript(self, script) --> None
        
        Sets the script of the current model. This overwrite the current 
        script.
        '''

        self.parent.script_editor.SetText(script)
        self.parent.model.set_script(script)

    def GetModelScript(self):
        """Returns the model script"""
        return self.parent.model.get_script()

    def CompileScript(self):
        '''CompileScript(self) --> None
        
        Compiles the model script
        '''
        self.parent.model.compile_script()

    def GetScriptModule(self):
        """Returns the script module"""
        return self.parent.model.script_module

    def OnNewModel(self, event):
        '''OnNewModel(self) --> None
        
        Function to be overridden. Called when a new model is being created.
        '''
        pass

    def OnDataChanged(self, event):
        '''OnDataChanged(self) --> None
        
        Function to be overridden. Called when a new data set has been loaded
        or deleted.
        '''
        pass

    def OnOpenModel(self, event):
        '''OnOpenModel(self, event) --> None
        
        Function that is called after a new model has been loaded.
        Used to set up plugin specific model stuff. To be overridden
        '''
        pass

    def OnSimulate(self, event):
        '''OnSimulate(self, event) --> None
        
        Function that is called after a simulation has been done.
        To be overridden
        '''
        pass

    def OnFittingUpdate(self, event):
        '''OnFittingUpdate(self, event) --> None
        
        Function that is called when the fitting algorithm pushes an update event.
        To be overridden
        '''
        pass

    def OnGridChange(self, event):
        """Responds to a change in the grid.

        :param event:
        :return:
        """
        pass

    def Remove(self):
        '''Remove(self) --> None
        Removes all components.
        '''
        pnb=self.parent.plot_notebook
        inb=self.parent.input_notebook
        dnb=self.parent.data_notebook

        # remove all pages from the notebooks
        for panel in self.plot_pages:
            pages=[pnb.GetPage(i) for i in range(pnb.GetPageCount())]
            idx=pages.index(panel)
            pnb.DeletePage(idx)
        for panel in self.input_pages:
            pages=[inb.GetPage(i) for i in range(inb.GetPageCount())]
            idx=pages.index(panel)
            inb.DeletePage(idx)
        for panel in self.data_pages:
            pages=[dnb.GetPage(i) for i in range(dnb.GetPageCount())]
            idx=pages.index(panel)
            dnb.DeletePage(idx)
        # Remove the menus
        for name in self.menus:
            idx=self.parent.main_frame_menubar.FindMenu(name)
            self.parent.main_frame_menubar.Remove(idx)

@dataclass
class PluginConfig(BaseConfig):
    section='plugins'
    loaded_plugins: str=''

class PluginController(Configurable):
    '''
    A controller class to interact with the gui
    so we can load and unload modules as well as
    update the module list.
    '''
    opt: PluginConfig

    def __init__(self, parent, menu):
        '''
        Insert menu items for controlling plugins in menu.
        Parent is the main window.
        '''
        self.plugin_handler=PluginHandler(parent, __MODULE_DIR__, 'add_ons')
        self.parent=parent
        Configurable.__init__(self)

        # make the menus
        self.load_menu=wx.Menu()
        menu.Insert(0, -1, 'Load', self.load_menu, 'Load a plugin')
        self.unload_menu=wx.Menu()
        menu.Insert(1, -1, 'Unload', self.unload_menu, 'Unload a plugin')

        menu.Append(-1, 'Update module list')

    def update_plugins(self):
        '''update_modules(self) --> None
        
        Updates the list of modules that can be loaded.
        '''
        # Remove all the items in load_menu
        items=self.load_menu.GetMenuItems()
        [self.load_menu.Delete(item) for item in items]

        # Get the new list of plugin modules
        modlist=self.plugin_handler.get_plugins()
        modlist.sort()

        # Add new menu items
        for mod in modlist:
            menu=self.load_menu.Append(-1, mod)
            self.parent.Bind(wx.EVT_MENU, self.LoadPlugin, menu)

        self.update_config()

    def RegisterPlugin(self, plugin):
        ''' RegisterPlugin(self, plugin) --> None
        
        Adds a plugin to the unload list so that it can be removed later.
        '''
        menu=self.unload_menu.Append(-1, plugin)
        self.parent.Bind(wx.EVT_MENU, self.UnLoadPlugin, menu)
        self.update_plugins()

    def update_config(self):
        '''update_config(self) --> None
        
        Updates the config object
        '''
        loaded_plugins=self.plugin_handler.get_loaded_plugins()
        self.opt.loaded_plugins=';'.join(loaded_plugins)
        self.WriteConfig()

    def LoadDefaultPlugins(self):
        '''LoadDefaultPlugins(self) --> None
        
        Tries to load the default plugins from the config object
        if they are not already loaded.
        '''
        self.ReadConfig()
        plugin_str=self.opt.loaded_plugins

        if plugin_str=='':
            self.update_plugins()
            return
        existing_plugins=self.plugin_handler.get_possible_plugins()

        for plugin in plugin_str.split(';'):
            # Check so the plugin is not loaded and exists 
            if not self.plugin_handler.is_loaded(plugin):
                if plugin in existing_plugins:
                    try:
                        self.plugin_handler.load_plugin(plugin)
                        self.RegisterPlugin(plugin)
                    except:
                        outp=io.StringIO()
                        traceback.print_exc(200, outp)
                        tbtext=outp.getvalue()
                        outp.close()
                        ShowErrorDialog(self.parent, 'Can NOT load plugin ' \
                                        +plugin+'\nPython traceback below:\n\n'+tbtext)
                        self.RegisterPlugin(plugin)
                else:
                    ShowInfoDialog(self.parent, 'Could not find plugin "%s"' \
                                                '. Either there is an error in the config file' \
                                                ' or the plugin is not installed.'%plugin)
        self.update_plugins()

    # Callbacks
    def LoadPlugin(self, event):
        '''OnLoadPlugin(self, event) --> None
        
        Loads a plugin from a menu choice.
        '''
        # Get the name of the plugin
        menuitem=self.load_menu.FindItemById(event.GetId())
        plugin=menuitem.GetItemLabel()
        try:
            self.plugin_handler.load_plugin(plugin)
        except:
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            tbtext=outp.getvalue()
            outp.close()
            ShowErrorDialog(self.parent, 'Can NOT load plugin '+plugin \
                            +'\nPython traceback below:\n\n'+tbtext)
        else:
            self.RegisterPlugin(plugin)

    def UnLoadPlugin(self, event):
        '''UnLoadPlugin(self, event) --> None
        
        UnLoads (removes) a plugin module.
        '''
        menuitem=self.unload_menu.FindItemById(event.GetId())
        plugin=menuitem.GetItemLabel()
        if self.UnLoadPlugin_by_Name(plugin):
            # Remove the item from the list
            self.unload_menu.Delete(menuitem)
            # Update the available plugins
            self.update_plugins()

    def UnLoadPlugin_by_Name(self, plugin):
        try:
            self.plugin_handler.unload_plugin(plugin)
        except Exception as e:
            outp=io.StringIO()
            traceback.print_exc(200, outp)
            tbtext=outp.getvalue()
            outp.close()
            ShowErrorDialog(self.parent, 'Can NOT unload plugin object'+ \
                            plugin+'\nPython traceback below:\n\n'+tbtext)
            return False
        else:
            return True

    def OnNewModel(self, event):
        '''OnNewModel(self, event) --> None
        
        Runs plugin code when the user tries to load a new model 
        '''
        for name in self.plugin_handler.loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnNewModel(event)

    def OnDataChanged(self, event):
        '''OnNewModel(self, event) --> None
        
        Runs plugin code when the user tries to load new data
        '''
        for name in self.plugin_handler.loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnDataChanged(event)

    def OnOpenModel(self, event):
        '''OnOpenModel(self, event) --> None
        
        Runs plugin code when the user tries to open a model 
        '''
        loaded_plugins=list(self.plugin_handler.loaded_plugins.keys())
        items=self.unload_menu.GetMenuItems()
        for item in items:
            # Remove the item from the list
            self.unload_menu.Delete(item)

        for name in loaded_plugins:
            # self.plugin_handler.loaded_plugins[name].OnOpenModel(event)
            self.plugin_handler.unload_plugin(name)

        self.LoadDefaultPlugins()
        # Update the available plugins
        self.update_plugins()

        loaded_plugins=list(self.plugin_handler.loaded_plugins.keys())
        for name in loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnOpenModel(event)

    def OnSimulate(self, event):
        '''OnOpenModel(self, event) --> None
        
        Runs plugin code when the user has simulated the model
        '''
        for name in self.plugin_handler.loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnSimulate(event)

    def OnFittingUpdate(self, event):
        '''OnOpenModel(self, event) --> None
        
        Runs plugin code when the Fitting algorithm wants to update its output.
        '''
        for name in self.plugin_handler.loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnFittingUpdate(event)

    def OnGridChanged(self, event):
        """ Runs plugin code to respond to a changed grid.

        :param event:
        :return:
        """
        for name in self.plugin_handler.loaded_plugins:
            self.plugin_handler.loaded_plugins[name].OnGridChange(event)

# ==============================================================================
# Utility Dialog functions..
def ShowInfoDialog(frame, message):
    dlg=wx.MessageDialog(frame, message,
                         'Information',
                         wx.OK | wx.ICON_INFORMATION
                         )
    dlg.ShowModal()
    dlg.Destroy()

def ShowErrorDialog(frame, message, position=''):
    dlg=wx.MessageDialog(frame, message,
                         'ERROR',
                         wx.OK | wx.ICON_ERROR
                         )
    dlg.ShowModal()
    dlg.Destroy()

def ShowWarningDialog(frame, message):
    dlg=wx.MessageDialog(frame, message, 'Warning',
                         wx.OK | wx.ICON_ERROR
                         )
    dlg.ShowModal()
    dlg.Destroy()

def ShowQuestionDialog(frame, message, title='Question'):
    dlg=wx.MessageDialog(frame, message,
                         title,
                         wx.YES_NO | wx.ICON_QUESTION
                         )
    result=dlg.ShowModal()==wx.ID_YES
    dlg.Destroy()
    return result



"""
Module to handle different cases of exceptions in the user interface.
"""

import traceback
from logging import debug, error

import wx

from ..exceptions import GenxError

class CatchModelError:
    """
    A context manager that displays an error dialog if an exception is raised within the 'with' block.
    If the error is a GenX exception the display is slightly different.
    """
    def __init__(self, parent:wx.TopLevelWindow=None,
                 action:str='execution', step:str=None,
                 status_update: callable=None):
        self.parent=parent
        self.action=action
        self.step=step
        self._status_update=status_update
        self.successful=False

    def status_update(self, text):
        if self._status_update:
            self._status_update(text)

    def __enter__(self):
        debug(f'enter {self.action}/{self.step}')
        if self.step:
            self.status_update(f'Start {self.step}.')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            debug(f'exit {self.action}/{self.step} w/o error')
            self.successful=True
            self.status_update(f'Success in {self.step}.')
            return True

        # allow the code to do something with the error, normally not needed
        self.successful=False
        self.exc_type=exc_type
        self.exc_val=exc_val
        self.exc_tb=exc_tb

        message=f"{type(exc_val).__name__} in {self.action}"
        if self.step:
            message+=f", could not {self.step}."
        ext_message=f"{exc_val}"

        if isinstance(exc_val, GenxError):
            self.status_update(f'Error in {self.step}, {type(exc_val).__name__}')
            title='Warning'
            icon_style=wx.ICON_WARNING
        else:
            self.status_update(f'Fatal error in {self.step}, {type(exc_val).__name__}')
            ext_message+='\n\nPython Error (last calls first):\n    '
            ext_message+='\n    '.join(traceback.format_tb(exc_tb)[:2:-1])
            title='Warning'
            icon_style=wx.ICON_ERROR
        full_trace = message+':\n\n'
        full_trace += ''.join(traceback.format_tb(exc_tb))
        full_trace += f'{type(exc_val).__name__}: {exc_val}'
        error(full_trace)

        # make sure the dialog is shown from main thread and after any queued actions
        wx.CallAfter(self.display_message, title, message, ext_message, full_trace, icon_style)
        return True # exception is not raised in main context

    def display_message(self, title, message, ext_message, full_trace, icon_style=wx.ICON_ERROR):
        style=wx.OK|wx.HELP
        dlg = wx.MessageDialog(self.parent, message, title, style|icon_style)
        dlg.SetExtendedMessage(ext_message)
        dlg.SetHelpLabel('Copy to Clipboard')

        result=dlg.ShowModal()
        while result==wx.ID_HELP:
            if wx.TheClipboard.Open():
                wx.TheClipboard.SetData(wx.TextDataObject(full_trace))
                wx.TheClipboard.Close()
            result=dlg.ShowModal()
        dlg.Destroy()

